"""
finding and digesting potential primer regions.
"""

# varVAMP
from varvamp.scripts import config


def find_regions(consensus_amb, allowed_ambiguous):
    """
    finds primer regions as specified by a
    certain amount of ambiguous bases in a given
    sequence length
    """

    current_window, primer_regions, all_previous_ambiguous_pos, last_amb = [], [], [], 0

    # append one N so the last window is closed
    seq = str(consensus_amb) + "N"

    for idx, nuc in enumerate(seq):
        # open a new window if none is there and we have a nucleotide
        if nuc in config.NUCS and not current_window:
            current_window = [idx, 0]
            all_previous_ambiguous_pos = []
        # if we have a non-nucleotide, check if its time to writ the window
        elif nuc not in config.NUCS and current_window:
            # are the current and last ambig char sufficiently far?
            if idx - last_amb >= config.PRIMER_SIZES[0] and nuc != "N":
                current_window[1] = 1
                all_previous_ambiguous_pos = [idx]
            else:
                current_window[1] += 1
                all_previous_ambiguous_pos.append(idx)
            # check if there were too many previous amb char in subwindow
            # or if N is reached -> writeable
            if current_window[1] > allowed_ambiguous or nuc == "N":
                # check if the writable window has a sufficient length.
                if idx-current_window[0] >= config.PRIMER_SIZES[0]:
                    primer_regions.append([current_window[0], idx])
                # open new window if N is reached
                if nuc == "N":
                    current_window = []
                # or reset to a previous ambig char
                else:
                    current_window[0] = all_previous_ambiguous_pos[0]+1
                    current_window[1] = current_window[1] - 1
                    all_previous_ambiguous_pos.pop(0)
            last_amb = idx

    return primer_regions


def mean(primer_regions, consensus):
    """
    calculate the percentage of regions
    that cover the consensus sequence
    """
    covered_set = set()

    for region in primer_regions:
        pos_list = list(range(region[0], region[1]))
        [covered_set.add(x) for x in pos_list]

    return round(len(covered_set)/len(consensus)*100, 1)


def digest_seq(seq, kmer_size):
    """
    digest the sequence into kmers
    """
    return [[seq[i:i+kmer_size], i, i+len(seq[i:i+kmer_size])] for i in range(len(seq)-kmer_size+1)]


def produce_kmers(primer_regions, consensus, sizes=config.PRIMER_SIZES):
    """
    produce kmers for all primer regions
    """
    kmers = set()

    for region in primer_regions:
        sliced_seq = consensus[region[0]:region[1]]
        for kmer_size in range(sizes[0], sizes[1]+1):
            kmers_temp = digest_seq(sliced_seq, kmer_size)
            # adjust the start and stop position of the kmers
            for kmer_temp in kmers_temp:
                kmer_temp[1] = kmer_temp[1]+region[0]
                kmer_temp[2] = kmer_temp[2]+region[0]
                # check if kmer is already in list (overlapping regions)
                kmers.add(tuple(kmer_temp))

    return kmers



"""
data writing and visualization.
"""
# BUILT-INS
import os
import math

# LIBS
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


# varVAMP
from varvamp.scripts import primers
from varvamp.scripts import config
from varvamp.scripts import logging


def write_fasta(path, file_name, seq_id, seq):
    """
    write fasta files
    """
    name = f"{file_name}.fasta"
    out = os.path.join(path, name)
    with open(out, 'w') as o:
        print(f">{seq_id}\n{seq}", file=o)


def write_alignment(path, alignment):
    """
    write alignment to file
    """
    name = "alignment_cleaned.fasta"
    out = os.path.join(path, name)
    with open(out, "w") as o:
        for seq in alignment:
            print(f">{seq[0]}\n{seq[1]}", file=o)


def write_regions_to_bed(primer_regions, scheme_name, path, mode=None):
    """
    write primer regions as bed file
    """

    if mode == "probe":
        outfile = f"{path}probe_regions.bed"
    else:
        outfile = f"{path}primer_regions.bed"

    with open(outfile, 'w') as o:
        for counter, region in enumerate(primer_regions):
            print(
                f"{scheme_name}_ambiguous_consensus",
                region[0],
                region[1],
                "REGION_"+str(counter),
                sep="\t",
                file=o
            )


def write_primers_to_bed(outfile, scheme_name, primer_name, primer_properties, numeric_value, direction, sequence=None):
    """
    write primers as bed file
    """
    with open(outfile, 'a') as o:
        # write header for primer bed
        data = [f"{scheme_name}_ambiguous_consensus",
            primer_properties[1],  # start
            primer_properties[2],  # stop
            primer_name,
            numeric_value,  # can be pool or score
            direction]
        if sequence is not None:
            data.append(sequence)
        print(
            *data,
            sep="\t",
            file=o
        )


def write_all_primers(path, scheme_name, all_primers):
    """
    write all primers that varVAMP designed as bed file
    """
    outfile = f"{path}all_primers.bed"

    for direction in all_primers:
        for primer in all_primers[direction]:
            write_primers_to_bed(outfile, scheme_name, primer, all_primers[direction][primer], round(all_primers[direction][primer][3], 2), direction)




def calc_mean_stats(permutations):
    """
    calculate mean gc and temp over all permutations
    """
    gc = 0
    temp = 0

    for permutation in permutations:
        gc += primers.calc_gc(permutation)
        temp += primers.calc_temp(permutation)

    return round(gc/len(permutations), 1), round(temp/len(permutations), 1)


def write_qpcr_to_files(path, final_schemes, ambiguous_consensus, scheme_name, log_file):
    """
    write all relevant bed files and tsv file for the qPCR design
    """

    tsv_file = os.path.join(path, "qpcr_design.tsv")
    tsv_file_2 = os.path.join(path, "qpcr_primers.tsv")
    primer_bed_file = os.path.join(path, "primers.bed")
    amplicon_bed_file = os.path.join(path, "amplicons.bed")
    primer_fasta_file = os.path.join(path, "oligos.fasta")

    with open(tsv_file, "w") as tsv, open(tsv_file_2, "w") as tsv2, open(amplicon_bed_file, "w") as bed, open(primer_fasta_file, "w") as fasta:
        print(
            "qpcr_scheme\toligo_type\tstart\tstop\tseq\tsize\tgc_best\ttemp_best\tmean_gc\tmean_temp\tpenalty\toff_target_amplicons",
            file=tsv2
        )
        print(
            "qpcr_scheme\toff_target_amplicons\tpenalty\tdeltaG\tlength\tstart\tstop\tseq",
            file=tsv
        )
        for n, amp in enumerate(final_schemes):
            amp_name = f"{scheme_name}_{n}"
            # write bed amplicon file
            print(
                f"{scheme_name}_ambiguous_consensus",
                amp["LEFT"][1],
                amp["RIGHT"][2],
                amp_name,
                round(amp["penalty"], 1),
                ".",
                sep="\t",
                file=bed
            )
            # write tsv
            amplicon_start = amp["LEFT"][1]
            amplicon_stop = amp["RIGHT"][2]
            if "off_targets" in amp:
                if amp["off_targets"]:
                    amplicon_has_off_target = "Yes"
                    write_BLAST_warning(amp_name, log_file)
                else:
                    amplicon_has_off_target = "No"
            else:
                amplicon_has_off_target = "n.d."
            amplicon_seq = ambiguous_consensus[amplicon_start:amplicon_stop]
            print(
                amp_name,
                amplicon_has_off_target,
                round(amp["penalty"], 1),
                amp["deltaG"],
                len(amplicon_seq),
                amplicon_start + 1,
                amplicon_stop,
                amplicon_seq,
                sep="\t",
                file=tsv
            )
            # write tsv2
            for oligo_type in ["LEFT", "PROBE", "RIGHT"]:
                seq = ambiguous_consensus[amp[oligo_type][1]:amp[oligo_type][2]]
                if oligo_type == "RIGHT" or (oligo_type == "PROBE" and amp["PROBE"][5] == "-"):
                    seq = primers.rev_complement(seq)
                    direction = "-"
                else:
                    direction = "+"

                permutations = primers.get_permutations(seq)
                gc, temp = calc_mean_stats(permutations)
                primer_name = f"{amp_name}_{oligo_type}"

                print(
                    amp_name,
                    oligo_type,
                    amp[oligo_type][1] + 1,
                    amp[oligo_type][2],
                    seq.upper(),
                    len(seq),
                    round(primers.calc_gc(amp[oligo_type][0]), 1),
                    round(primers.calc_temp(amp[oligo_type][0]), 1),
                    gc,
                    temp,
                    round(amp[oligo_type][3], 1),
                    amplicon_has_off_target,
                    sep="\t",
                    file=tsv2
                )
                # write primer bed file
                write_primers_to_bed(
                    primer_bed_file,
                    scheme_name,
                    primer_name,
                    amp[oligo_type],
                    round(amp[oligo_type][3], 2),
                    direction,
                    seq.upper()
                )
                # write fasta
                print(f">{primer_name}\n{seq.upper()}", file=fasta)


def write_scheme_to_files(path, amplicon_scheme, ambiguous_consensus, scheme_name, mode, log_file, primer_dimers=None):
    """
    write all relevant bed files and a tsv file with all primer stats
    """
    tsv_file = os.path.join(path, "primers.tsv")
    primer_bed_file = os.path.join(path, "primers.bed")
    amplicon_bed_file = os.path.join(path, "amplicons.bed")
    tabular_file = os.path.join(path, "primer_to_amplicon_assignment.tabular")

    # Map old primer names to new amplicon-based names
    name_mapping = {}

    # open files to write
    with open(tsv_file, "w") as tsv, open(amplicon_bed_file, "w") as bed, open(tabular_file, "w") as tabular:
        # write header for primer tsv
        print(
            "amlicon_name\tamplicon_length\tprimer_name\tprimer_name_all_primers\tpool\tstart\tstop\tseq\tsize\tgc_best\ttemp_best\tmean_gc\tmean_temp\tpenalty\toff_target_amplicons",
            file=tsv
        )
        amplicon_bed_records = []
        primer_bed_records = []
        primer_assignment_records = []
        pools = {amp.get("pool", 0) for amp in amplicon_scheme}
        for pool in pools:
            if mode == "single":
                primer_fasta_file = os.path.join(path, "primers.fasta")
            else:
                primer_fasta_file = os.path.join(path, f"primers_pool_{pool + 1}.fasta")
            with open(primer_fasta_file, "w") as primer_fasta:
                for counter, amp in enumerate(amplicon_scheme[pool::len(pools)]):
                    # give a new amplicon name
                    amplicon_index = counter * len(pools) + pool
                    amp_name = f"{scheme_name}_{amplicon_index}"
                    # get left and right primers and their names
                    amp_length = amp["RIGHT"][2] - amp["LEFT"][1]
                    if "off_targets" in amp:
                        if amp["off_targets"]:
                            amplicon_has_off_target = "Yes"
                            write_BLAST_warning(amp_name, log_file)
                        else:
                            amplicon_has_off_target = "No"
                    else:
                        amplicon_has_off_target = "n.d."
                    # write amplicon bed
                    if mode == "tiled":
                        bed_score = pool + 1
                    elif mode == "single":
                        bed_score = round(amp["LEFT"][3] + amp["RIGHT"][3], 1)
                    amplicon_bed_records.append(
                        (
                            amp["LEFT"][1],
                            amp["RIGHT"][2],
                            amp_name,
                            bed_score
                        )
                    )
                    primer_assignment_records.append(
                        (
                            # will need amplicon_index for sorting
                            amplicon_index,
                            (f"{amp_name}_LEFT", f"{amp_name}_RIGHT")
                        )
                    )
                    # Build name mapping for dimers
                    name_mapping[amp["LEFT"][-1]] = f"{amp_name}_LEFT"
                    name_mapping[amp["RIGHT"][-1]] = f"{amp_name}_RIGHT"

                    # write primer tsv and primer bed
                    for direction, primer in [("+", amp["LEFT"]), ("-", amp["RIGHT"])]:
                        seq = ambiguous_consensus[primer[1]:primer[2]]
                        if direction == "-":
                            seq = primers.rev_complement(seq)
                            primer_name = f"{amp_name}_RIGHT"
                        else:
                            primer_name = f"{amp_name}_LEFT"
                        # write primers to fasta pool file
                        print(f">{primer_name}\n{seq.upper()}", file=primer_fasta)
                        # calc primer parameters for all permutations
                        permutations = primers.get_permutations(seq)
                        gc, temp = calc_mean_stats(permutations)
                        # write tsv file
                        print(
                            amp_name,
                            amp_length,
                            primer_name,
                            primer[-1],
                            pool + 1,
                            primer[1] + 1,
                            primer[2],
                            seq.upper(),
                            len(primer[0]),
                            round(primers.calc_gc(primer[0]), 1),
                            round(primers.calc_temp(primer[0]), 1),
                            gc,
                            temp,
                            round(primer[3], 1),
                            amplicon_has_off_target,
                            sep="\t",
                            file=tsv
                        )
                        primer_bed_records.append(
                            (
                                # will need amplicon_index for sorting
                                amplicon_index,
                                (primer_name, primer, pool + 1, direction, seq.upper())
                            )
                        )
        # write amplicon bed with amplicons sorted by start position
        for record in sorted(amplicon_bed_records, key=lambda x: x[0]):
            print(
                f"{scheme_name}_ambiguous_consensus",
                *record,
                ".",
                sep="\t",
                file=bed
            )
        # use sorting by amplicon index for primer assignment file
        for record in sorted(primer_assignment_records):
            print(
                *record[1],
                sep="\t",
                file=tabular
            )
        # same for primer bed
        for record in sorted(primer_bed_records):
            write_primers_to_bed(
                primer_bed_file,
                scheme_name,
                *record[1]
            )

    # Write dimers with renamed primers
    if primer_dimers:
        write_dimers(path, primer_dimers, name_mapping)


def write_dimers(path, primer_dimers, name_mapping):
    """
    write dimers for which no replacement was found to file
    """
    file = os.path.join(path, "unsolvable_primer_dimers.txt")
    with open(file, "w") as f:
        for pool, primer1, primer2 in primer_dimers:
            dimer_result = primers.calc_dimer(primer1[2][0], primer2[2][0], structure=True)
            print(
                "pool\tprimer 1\tprimer 2\tdimer melting temp\tdeltaG",
                file=f
            )
            print(
                pool+1,
                name_mapping[primer1[1]],
                name_mapping[primer2[1]],
                round(dimer_result.tm, 1),
                dimer_result.dg,
                sep="\t",
                file=f
            )
            structure = [x[4:] for x in dimer_result.ascii_structure_lines]
            print("\nDimer structure:", file=f)
            for line in structure:
                print(
                    line,
                    file=f
                )
            print(file=f)


def entropy(chars, states):
    """
    input is a list of characters or numbers.
    calculate relative shannon's entropy. relative values are
    achieved by using the number of possible states as the base
    """
    ent = 0
    n_chars = len(chars)
    # only one char is in the list
    if n_chars <= 1:
        return ent
    # calculate the number of unique chars and their counts
    value, counts = np.unique(chars, return_counts=True)
    probs = counts/n_chars
    if np.count_nonzero(probs) <= 1:
        return ent

    for prob in probs:
        ent -= prob*math.log(prob, states)

    return ent

def alignment_entropy(alignment_cleaned):
    """
    calculate the entropy for every position in an alignment.
    return pandas df.
    """
    position = list()
    entropys = list()
    # iterate over alignment positions and the sequences
    for nuc_pos in range(0, len(alignment_cleaned[0][1])):
        pos = []
        for seq_number in range(0, len(alignment_cleaned)):
            pos.append(alignment_cleaned[seq_number][1][nuc_pos])
        entropys.append(entropy(pos, 4))
        position.append(nuc_pos)
    # create df
    entropy_df = pd.DataFrame()
    entropy_df["position"] = position
    entropy_df["entropy"] = entropys
    entropy_df["average"] = entropy_df["entropy"].rolling(10, center=True).mean()

    return entropy_df


def entropy_subplot(ax, alignment_cleaned, scheme_name):
    """
    creates the entropy subplot
    """
    # - create entropy df
    entropy_df = alignment_entropy(alignment_cleaned)

    ax[0].fill_between(entropy_df["position"], entropy_df["entropy"], color="gainsboro", label="entropy")
    ax[0].plot(entropy_df["position"], entropy_df["average"], color="black", label="average entropy", linewidth=0.5)
    ax[0].set_ylim((0, 1))
    ax[0].set_xlim(0, max(entropy_df["position"]))
    ax[0].set_ylabel("normalized Shannon's entropy")
    ax[0].set_title(f"{scheme_name} amplicon design")
    ax[0].spines['top'].set_visible(False)
    ax[0].spines['right'].set_visible(False)


def region_subplot(ax, primer_regions, location=0.95, color="darkorange", description="possible primer regions"):
    """
    creates the primer regions subplot
    """
    for region in primer_regions:
        ax[1].hlines(location, region[0], region[1], linewidth=5, color=color)
    # legend
    ax[1].hlines(location, primer_regions[0][1], primer_regions[0][1], label=description, linewidth=5, color=color)


def all_primer_subplot(ax, all_primers):
    """
    creates the all primer subplot
    """
    for direction in all_primers:
        if direction == "-":
            primer_position = 0.85
            primer_color = "darkgrey"
            primer_label = "all right primers"
        elif direction == "+":
            primer_position = 0.8
            primer_color = "dimgrey"
            primer_label = "all left primers"
        for primer in all_primers[direction]:
            ax[1].hlines(primer_position, all_primers[direction][primer][1], all_primers[direction][primer][2], linewidth=5, color=primer_color)
        # legend
        ax[1].hlines(primer_position, all_primers[direction][primer][1], all_primers[direction][primer][2], linewidth=5, color=primer_color, label=primer_label)


def amplicon_subplot(ax, amplicon_scheme):
    """
    creates the amplicon subplot
    """
    for counter, amp in enumerate(amplicon_scheme):
        pool = amp.get("pool", 0)
        if pool == 0:
            position_amp = 0.7
            position_text = 0.6
        elif pool == 1:
            position_amp = 0.6
            position_text = 0.65
        left = amp["LEFT"]
        right = amp["RIGHT"]
        # amplicons
        ax[1].hlines(position_amp, left[1], right[2], linewidth=5)
        # text
        ax[1].text(right[2] - (right[2]-left[1])/2, position_text, str(counter), fontsize=8)
        # primers
        ax[1].hlines(position_amp, left[1], left[2], linewidth=5, color="red")
        ax[1].hlines(position_amp, right[1], right[2], linewidth=5, color="red")
    # legends
    ax[1].hlines(position_amp, left[1]+config.PRIMER_SIZES[1], right[2]-config.PRIMER_SIZES[1], linewidth=5, label="amplicons")
    ax[1].hlines(position_amp, left[1], left[2], linewidth=5, color="red", label="primers")


def qpcr_subplot(ax, amplicon_scheme):
    """
    creates the qpcr subplot
    """
    for counter, amp in enumerate(amplicon_scheme):
        left = amp["LEFT"]
        right = amp["RIGHT"]
        probe = amp["PROBE"]
        # amplicons
        ax[1].hlines(0.8, left[1], right[2], linewidth=5)
        # text
        ax[1].text(right[2] - (right[2]-left[1])/2, 0.65, str(counter), fontsize=8)
        # primers
        ax[1].hlines(0.8, left[1], left[2], linewidth=5, color="red")
        ax[1].hlines(0.8, right[1], right[2], linewidth=5, color="red")
        # probe
        ax[1].hlines(0.75, probe[1], probe[2], linewidth=5, color="darkgrey")

    # legends
    ax[1].hlines(0.8, left[1]+config.PRIMER_SIZES[1], right[2]-config.PRIMER_SIZES[1], linewidth=5, label="amplicons")
    ax[1].hlines(0.8, left[1], left[2], linewidth=5, color="red", label="primers")
    ax[1].hlines(0.75, probe[1], probe[2], linewidth=5, color="darkgrey", label="probe")


def varvamp_plot(path, alignment_cleaned, primer_regions, scheme_name, all_primers=None, amplicon_scheme=None, probe_regions=None):
    """
    creates overview plot for the amplicon design
    and per base coverage plots
    """
    # first plot: overview
    # create pdf name
    name = "amplicon_plot.pdf"
    out = os.path.join(path, name)
    # ini figure
    fig, ax = plt.subplots(2, 1, figsize=[22, 6], squeeze=True, sharex=True, gridspec_kw={'height_ratios': [4, 1]})
    fig.subplots_adjust(hspace=0)
    # entropy plot
    entropy_subplot(ax, alignment_cleaned, scheme_name)
    # primer regions plot
    region_subplot(ax, primer_regions)
    # probe region plot for probes
    if probe_regions is not None and amplicon_scheme is not None:
        region_subplot(ax, probe_regions, 0.9, color="dimgrey", description="possible probe regions")
        qpcr_subplot(ax, amplicon_scheme)
    # all primer plot
    elif all_primers is not None and amplicon_scheme is not None:
        all_primer_subplot(ax, all_primers)
        # amplicon, text and primer plot
        amplicon_subplot(ax, amplicon_scheme)
    # finalize
    ax[1].spines['right'].set_visible(False)
    ax[1].spines['left'].set_visible(False)
    ax[1].spines['bottom'].set_visible(False)
    ax[1].axes.get_yaxis().set_visible(False)
    ax[1].set_xlabel("alignment position")
    ax[1].set_ylim((0.5, 1))
    fig.legend(loc=(0.83, 0.7))

    # save fig
    fig.savefig(out, bbox_inches='tight')
    plt.close()


def get_primers_for_plot(amplicon_scheme, scheme_name, mode):
    """
    get the primers for per base pair plot (single, tiled)
    """
    amplicon_primers = []

    if mode == "SINGLE/TILED":
        oligo_types = ["LEFT", "RIGHT"]
    else:
        oligo_types = ["PROBE", "LEFT", "RIGHT"]

    for counter, amp in enumerate(amplicon_scheme):
        for oligo_type in oligo_types:
            primer_name = f"{scheme_name}_{counter}_{oligo_type}"
            amplicon_primers.append((primer_name, amp[oligo_type]))

    return amplicon_primers


def per_base_mismatch_plot(path, amplicon_scheme, threshold, scheme_name, mode="SINGLE/TILED"):
    """
    per base pair mismatch multiplot
    """
    out = os.path.join(path, "per_base_mismatches.pdf")

    amplicon_primers = get_primers_for_plot(amplicon_scheme, scheme_name, mode)
    # ini multi pdf
    with PdfPages(out) as pdf:
        # always print 4 primers to one page
        for i in range(0, len(amplicon_primers), 4):
            # ini figure
            primers_temp = amplicon_primers[i:i+4]
            fig, ax = plt.subplots(len(primers_temp), figsize=(12, len(primers_temp)*4), squeeze=True)
            # edge case if primer_temp has the length 1
            if len(primers_temp) == 1:
                ax = [ax]
            fig.suptitle("Per base mismatches", fontsize=18)
            fig.tight_layout(rect=[0.05, 0.05, 1, 0.98])
            fig.subplots_adjust(hspace=0.5)
            # plotting
            for idx, primer in enumerate(primers_temp):
                x = [pos+primer[1][1] for pos in range(0, len(primer[1][4]))]
                ax[idx].bar(x, primer[1][4], color='lightgrey', edgecolor='black')
                ax[idx].set_title(primer[0], loc="left")
                ax[idx].xaxis.set_ticks(np.arange(primer[1][1], primer[1][1]+len(x), 1))
                ax[idx].xaxis.set_ticklabels(x, rotation=45)
                ax[idx].set_ylabel(ylabel="freq of sequences")
                ax[idx].set_xlabel("position")
                # set yaxis lims reasonably
                if threshold < 1:
                    ax[idx].set_ylim(0, 1-threshold)
                else:
                    ax[idx].set_ylim(0, 0.001)
            # - to pdf
            pdf.savefig(fig, bbox_inches='tight')
            plt.close()


def write_BLAST_warning(amplicon_name, log_file):
    """
    for each primer pair that has potential unspecific amplicons
    write warnings to file.
    """
    logging.raise_error(
        f"{amplicon_name} could produce off-targets. No better amplicon in this area was found.",
        log_file,
        exit=False,
    )



"""
primer creation and evaluation
"""

# BUILTIN
import itertools
import re
import multiprocessing
import functools

# LIBS
from Bio.Seq import MutableSeq
from Bio import SeqIO
import primer3 as p3

# varVAMP
from varvamp.scripts import config


def calc_gc(seq):
    """
    calculate the gc of a sequence
    """
    return 100*(seq.count("g")+seq.count("c"))/len(seq)


def calc_temp(seq):
    """
    calculate the melting temperature
    """
    return p3.calc_tm(
            seq.upper(),
            mv_conc=config.PCR_MV_CONC,
            dv_conc=config.PCR_DV_CONC,
            dntp_conc=config.PCR_DNTP_CONC,
            dna_conc=config.PCR_DNA_CONC
        )


def calc_hairpin(seq):
    """
    calculates hairpins
    """
    return p3.calc_hairpin(
            seq.upper(),
            mv_conc=config.PCR_MV_CONC,
            dv_conc=config.PCR_DV_CONC,
            dntp_conc=config.PCR_DNTP_CONC,
            dna_conc=config.PCR_DNA_CONC
        )


def calc_dimer(seq1, seq2, structure=False):
    """
    Calculate the hetero-dimerization thermodynamics of two DNA sequences.
    Return primer3 thermo object.
    """
    return p3.calc_heterodimer(
        seq1.upper(),
        seq2.upper(),
        mv_conc=config.PCR_MV_CONC,
        dv_conc=config.PCR_DV_CONC,
        dna_conc=config.PCR_DNA_CONC,
        dntp_conc=config.PCR_DNTP_CONC,
        output_structure=structure
    )


def has_end_overlap(dimer_result):
    """
    checks if two oligos overlap at their ends
    Example:
        xxxxxxxxtagc-------
        --------atcgxxxxxxx
    """
    if dimer_result.structure_found:
        # clean structure
        structure = [x[4:] for x in dimer_result.ascii_structure_lines]
        # check if we have an overlap that is large enough
        overlap = len(structure[1].replace(" ", ""))
        if overlap <= config.END_OVERLAP:
            return False
        # not more than one conseq. internal mismatch
        if '  ' in structure[1].lstrip(" "):
            return False
        # The alignment length of the ACII structure is equal to the first part of the structure
        # and the maximum possible alignment length is the cumulative length of both primers (-> no overlap at all)
        alignment_length = len(structure[0])
        maximum_alignment_length = len(re.findall("[ATCG]", "".join(structure)))
        # this means that for a perfect end overlap the alignment length is equal to:
        # len(primer1) + len(primer2) - overlap.
        if alignment_length == maximum_alignment_length - overlap:
            return True

    return False


def is_dimer(seq1, seq2):
    """
    check if two sequences dimerize above threshold or are overlapping at their ends
    """
    dimer_result = calc_dimer(seq1, seq2, structure=True)
    # check both the temperature and the deltaG
    if dimer_result.tm > config.PRIMER_MAX_DIMER_TMP or dimer_result.dg < config.PRIMER_MAX_DIMER_DELTAG:
        return True
    # check for perfect end overlaps (this can result in primer extensions even though the tm/dg are okay)
    if has_end_overlap(dimer_result):
        return True

    return False


def calc_max_polyx(seq):
    """
    calculate maximum polyx of a seq
    """
    previous_nuc, counter, max_polyx = seq[0], 1, 1

    for nuc in seq[1:]:
        if nuc == previous_nuc:
            counter += 1
        else:
            counter = 1
            previous_nuc = nuc
        if counter > max_polyx:
            max_polyx = counter

    return max_polyx


def calc_max_dinuc_repeats(seq):
    """
    calculate maximum polyxy of a seq
    """
    for s in [seq, seq[1:]]:
        previous_dinuc = s[0:2]
        max_dinuc = 1
        counter = 1
        for i in range(2, len(s), 2):
            if s[i:i+2] == previous_dinuc:
                counter += 1
            else:
                if counter > max_dinuc:
                    max_dinuc = counter
                counter = 1
                previous_dinuc = s[i:i+2]

    return max_dinuc


def calc_end_gc(seq):
    """
    check how many gc nucleotides
    are within the last 5 bases of
    the 3' end
    """
    return seq[-5:].count('g') + seq[-5:].count('c')


def is_three_prime_ambiguous(amb_seq):
    """
    determine if a sequence contains an ambiguous char at the 3'prime
    """
    len_3_prime, is_amb = config.PRIMER_MIN_3_WITHOUT_AMB, False

    if len_3_prime != 0:
        for nuc in amb_seq[len(amb_seq)-len_3_prime:]:
            if nuc not in config.NUCS:
                is_amb = True
                break

    return is_amb


def rev_complement(seq):
    """
    reverse complement a sequence
    """
    return str(MutableSeq(seq).reverse_complement(inplace=True))


def calc_permutation_penalty(amb_seq):
    """
    get all permutations of a primer with ambiguous
    nucleotides and multiply with permutation penalty
    """
    permutations = 0

    for nuc in amb_seq:
        if nuc in config.AMBIG_NUCS:
            n = len(config.AMBIG_NUCS[nuc])
            if permutations != 0:
                permutations = permutations*n
            else:
                permutations = n

    return permutations*config.PRIMER_PERMUTATION_PENALTY


def calc_base_penalty(seq, primer_temps, gc_range, primer_sizes):
    """
    Calculate intrinsic primer penalty.
    """
    penalty = 0

    tm = calc_temp(seq)
    gc = calc_gc(seq)
    size = len(seq)

    # TEMP penalty
    if tm > primer_temps[2]:
        penalty += config.PRIMER_TM_PENALTY*(
            tm - primer_temps[2]
            )
    if tm < primer_temps[2]:
        penalty += config.PRIMER_TM_PENALTY*(
            primer_temps[2] - tm
            )
    # GC penalty
    if gc > gc_range[2]:
        penalty += config.PRIMER_GC_PENALTY*(
            gc - gc_range[2]
            )
    if gc < gc_range[2]:
        penalty += config.PRIMER_GC_PENALTY*(
            gc_range[2] - gc
        )
    # SIZE penalty
    if size > primer_sizes[2]:
        penalty += config.PRIMER_SIZE_PENALTY*(
            size - primer_sizes[2]
        )
    if size < primer_sizes[2]:
        penalty += config.PRIMER_SIZE_PENALTY * (
            primer_sizes[2] - size
        )

    return penalty


def calc_per_base_mismatches(kmer, alignment, ambiguous_consensus):
    """
    calculate for a given kmer with [seq, start, stop]
    the percent mismatch per kmer pos with the alignment.
    considers if kmer or aln sequences have an amb nuc. returns
    a list of percent mismatches for each kmer position.
    """
    # ini list
    mismatches = len(kmer[0])*[0]
    # get kmer with ambiguous nucs
    amb_kmer = ambiguous_consensus[kmer[1]:kmer[2]]
    # test it against all sequences in the alignment
    for sequence in alignment:
        # slice each sequence of the alignment for the kmer
        # start and stop positions
        seq_slice = sequence[1][kmer[1]:kmer[2]]
        for idx, slice_nuc in enumerate(seq_slice):
            # find the respective nuc to that of the slice
            current_kmer_pos = amb_kmer[idx]
            if slice_nuc == current_kmer_pos:
                continue
            # check if the slice nucleotide is an amb pos
            if slice_nuc in config.AMBIG_NUCS:
                # check if the kmer has an amb pos
                if current_kmer_pos in config.AMBIG_NUCS:
                    slice_nuc_set = set(config.AMBIG_NUCS[slice_nuc])
                    pri_set = set(config.AMBIG_NUCS[current_kmer_pos])
                    # check if these sets have no overlap
                    # -> mismatch
                    if len(slice_nuc_set.intersection(pri_set)) == 0:
                        mismatches[idx] += 1
                # if no amb pos is in kmer then check if kmer nuc
                # is part of the amb slice nuc
                elif current_kmer_pos not in config.AMBIG_NUCS[slice_nuc]:
                    mismatches[idx] += 1
            # check if kmer has an amb pos but the current
            # slice_nuc is not part of this amb nucleotide
            elif current_kmer_pos in config.AMBIG_NUCS:
                if slice_nuc not in config.AMBIG_NUCS[current_kmer_pos]:
                    mismatches[idx] += 1
            # mismatch
            else:
                mismatches[idx] += 1

    # gives a percent mismatch over all positions of the kmer from 5' to 3'
    mismatches = [round(x/len(alignment), 2) for x in mismatches]

    return mismatches


def calc_3_prime_penalty(direction, mismatches):
    """
    calculate the penalty for mismatches at the 3' end.
    the more mismatches are closer to the 3' end of the kmer,
    the higher the penalty. uses the previously calculated
    mismatch list.
    """

    penalty = 0

    if config.PRIMER_3_PENALTY:
        if direction == "-":
            penalty = sum([m * p for m, p in zip(mismatches[0:len(config.PRIMER_3_PENALTY)], config.PRIMER_3_PENALTY)])
        elif direction == "+":
            penalty = sum([m * p for m, p in zip(mismatches[::-1][0:len(config.PRIMER_3_PENALTY)], config.PRIMER_3_PENALTY)])

    return penalty


def filter_kmer_direction_independent(seq, primer_temps=config.PRIMER_TMP, gc_range=config.PRIMER_GC_RANGE, primer_sizes=config.PRIMER_SIZES):
    """
    filter kmer for temperature, gc content,
    poly x, dinucleotide repeats and homodimerization
    """

    return(
        (primer_temps[0] <= calc_temp(seq) <= primer_temps[1])
        and (gc_range[0] <= calc_gc(seq) <= gc_range[1])
        and (calc_max_polyx(seq) <= config.PRIMER_MAX_POLYX)
        and (calc_max_dinuc_repeats(seq) <= config.PRIMER_MAX_DINUC_REPEATS)
        and (calc_base_penalty(seq, primer_temps, gc_range, primer_sizes) <= config.PRIMER_MAX_BASE_PENALTY)
        and not is_dimer(seq, seq)
    )


def filter_kmer_direction_dependend(direction, kmer, ambiguous_consensus):
    """
    filter for 3'ambiguous, hairpin temp and end GC content.
    this differs depending on the direction of the kmer.
    """
    # get the correct amb kmer to test
    if direction == "+":
        kmer_seq = kmer[0]
        amb_kmer_seq = ambiguous_consensus[kmer[1]:kmer[2]]
    elif direction == "-":
        kmer_seq = rev_complement(kmer[0])
        amb_kmer_seq = rev_complement(ambiguous_consensus[kmer[1]:kmer[2]])
    # filter kmer
    return (
        (calc_hairpin(kmer_seq).tm <= config.PRIMER_HAIRPIN)
        and (config.PRIMER_GC_END[0] <= calc_end_gc(kmer_seq) <= config.PRIMER_GC_END[1])
        and not is_three_prime_ambiguous(amb_kmer_seq)
    )


def _process_kmer_batch(ambiguous_consensus, alignment, kmers):
    """
    Helper function for multiprocessing: process a batch of kmers.
    Returns (left_primers, right_primers) tuples.
    """
    left_primers = []
    right_primers = []

    for kmer in kmers:
        if not filter_kmer_direction_independent(kmer[0]):
            continue
        # calc penalties
        base_penalty = calc_base_penalty(kmer[0], config.PRIMER_TMP, config.PRIMER_GC_RANGE, config.PRIMER_SIZES)
        per_base_mismatches = calc_per_base_mismatches(kmer, alignment, ambiguous_consensus)
        permutation_penalty = calc_permutation_penalty(ambiguous_consensus[kmer[1]:kmer[2]])
        # some filters depend on the direction of each primer
        for direction in ["+", "-"]:
            if not filter_kmer_direction_dependend(direction, kmer, ambiguous_consensus):
                continue
            # calc penalties
            three_prime_penalty = calc_3_prime_penalty(direction, per_base_mismatches)
            primer_penalty = base_penalty + permutation_penalty + three_prime_penalty
            # add to lists depending on their direction
            if direction == "+":
                left_primers.append([kmer[0], kmer[1], kmer[2], primer_penalty, per_base_mismatches])
            else:
                right_primers.append([rev_complement(kmer[0]), kmer[1], kmer[2], primer_penalty, per_base_mismatches])

    return left_primers, right_primers


def find_primers(kmers, ambiguous_consensus, alignment, num_processes):
    """
    Filter kmers direction specific and append penalties --> potential primers.
    Uses multiprocessing to process kmers in parallel.
    """
    if not kmers:
        return [], []

    # Convert kmers set to list for slicing
    kmers = list(kmers)
    batch_size = max(1, int(len(kmers)/num_processes))

    # Split kmers into batches
    batches = [kmers[i:i + batch_size] for i in range(0, len(kmers), batch_size)]
    callable_f = functools.partial(
        _process_kmer_batch,
        ambiguous_consensus, alignment
    )

    # Solve dimers in parallel
    with multiprocessing.Pool(processes=num_processes) as pool:
        results = pool.map(callable_f, batches)

    # Aggregate results
    left_primer_candidates = []
    right_primer_candidates = []
    for left_primers, right_primers in results:
        left_primer_candidates.extend(left_primers)
        right_primer_candidates.extend(right_primers)

    return left_primer_candidates, right_primer_candidates


def create_primer_dictionary(primer_candidates, direction):
    """
    creates a primer dictionary from primer list
    """
    primer_dict = {}
    primer_idx = 0

    for primer in primer_candidates:
        if direction == "+":
            direction_name = "LEFT"
        else:
            direction_name = "RIGHT"
        primer_name = f"{direction_name}_{primer_idx}"
        primer_dict[primer_name] = primer
        primer_idx += 1

    return primer_dict


def find_best_primers(left_primer_candidates, right_primer_candidates, high_conservation:bool=False):
    """
    Primer candidates are likely overlapping. Here, the list of primers
    is sorted for the lowest to highest penalty. Then, the next lowest
    is retained if it does not have any nucleotides that have already
    been covered by the middle third of a better primer candidate.
    This significantly reduces the amount of primers while retaining
    the ones with the lowest penalties.

    Example:
    -------- (penalty 1) 1
        ------------- (penalty 1) 2
            ------------ (penalty 0.8) 3
                      ----------(penalty 0.9) 4

    --> primer 3 would be retained and primer 2 excluded, primer 4 and 1
    will however be considered in the next set of overlapping primers.

    """
    all_primers = {}

    for direction, primer_candidates in [("+", left_primer_candidates), ("-", right_primer_candidates)]:
        # sort the primers by penalty, and if same penalty by start
        primer_candidates.sort(key=lambda x: (x[3], x[1]))
        # ini everything with the primer with the lowest penalty
        to_retain = [primer_candidates[0]]
        primer_set = set(range(primer_candidates[0][1], primer_candidates[0][2]))

        for primer in primer_candidates[1:]:
            # for highly conserved alignments exclude everything that overlaps with the best primer
            # this reduces graph complexity by quite a large margin
            if high_conservation:
                primer_positions =set(range(primer[1], primer[2]))
            # get the thirds of the primer, only consider the middle
            else:
                thirds_len = int((primer[2] - primer[1])/3)
                primer_positions = set(range(primer[1] + thirds_len, primer[2] - thirds_len))
            # check if none of the nucleotides of the next primer
            # are already covered by a better primer
            if primer_set.isdisjoint(primer_positions):
                # update the primer set
                primer_set.update(primer_positions)
                # append this primer as it has a low penalty and is not overlapping
                # with another already retained primer
                to_retain.append(primer)

        # sort by start
        to_retain.sort(key=lambda x: x[1])
        # create dict
        all_primers[direction] = create_primer_dictionary(to_retain, direction)

    # and create a dict
    return all_primers


def get_permutations(seq):
    """
    get all permutations of an ambiguous sequence.
    """
    splits = [config.AMBIG_NUCS.get(nuc, [nuc]) for nuc in seq]

    return[''.join(p) for p in itertools.product(*splits)]


def parse_primer_fasta(fasta_path):
    """
    Parse a primer FASTA file and return a list of sequences using BioPython.
    """

    sequences = []

    for record in SeqIO.parse(fasta_path, "fasta"):
        seq = str(record.seq).lower()
        # Only include primers up to 40 nucleotides
        if len(seq) <= 40:
            sequences += get_permutations(seq)

    return list(set(sequences))  # deduplication


def check_primer_against_externals(external_sequences, primer):
    """
    Worker function to check a single primer against all external sequences.
    Returns the primer if it passes, None otherwise.
    Handles both list format and dict format (name, data) tuples.
    """

    # Extract sequence based on input format
    if isinstance(primer, tuple):
        name, data = primer
        seq = data[0]
    else:
        seq = primer[0]

    for ext_seq in external_sequences:
        if is_dimer(seq, ext_seq):
            return None

    return primer


def filter_non_dimer_candidates(primer_candidates, external_sequences, n_processes):
    """
    Filter out primer candidates that form dimers with external sequences.
    Uses multiprocessing to speed up checks.
    """
    is_dict = isinstance(primer_candidates, dict)

    callable_f = functools.partial(
        check_primer_against_externals,
        external_sequences
    )

    with multiprocessing.Pool(processes=n_processes) as pool:
        # Prepare arguments based on input type
        # qpcr probes are stored in dictionaries --> result in tuples when unpacked
        if is_dict:
            results = pool.map(callable_f, primer_candidates.items())
        else:
            results = pool.map(callable_f, primer_candidates)

    # Filter and restore original format
    if is_dict:
        filtered_results = [result for result in results if result is not None]
        return {name: data for name, data in filtered_results}
    else:
        return [primer for primer in results if primer is not None]



"""
main workflow
"""

# BUILT-INS
import sys
import os
import datetime
import argparse

# varVAMP
from varvamp.scripts import alignment
from varvamp.scripts import config
from varvamp.scripts import consensus
from varvamp.scripts import regions
from varvamp.scripts import logging
from varvamp.scripts import param_estimation
from varvamp.scripts import primers
from varvamp.scripts import qpcr
from varvamp.scripts import reporting
from varvamp.scripts import scheme
from varvamp.scripts import blast
from varvamp import __version__


def get_args(sysargs):
    """
    arg parsing for varvamp
    """
    parser = argparse.ArgumentParser(
        usage='''\tvarvamp <mode> --help\n\tvarvamp <mode> [mode optional arguments] <alignment> <output dir>''')
    mode_parser = parser.add_subparsers(
        title="varvamp mode",
        dest="mode",
    )
    SINGLE_parser = mode_parser.add_parser(
        "single",
        help="design primers for single amplicons",
        usage="varvamp single [optional arguments] <alignment> <output dir>"
    )
    TILED_parser = mode_parser.add_parser(
        "tiled",
        help="design primers for whole genome sequencing",
        usage="varvamp tiled [optional arguments] <alignment> <output dir>"
    )
    QPCR_parser = mode_parser.add_parser(
        "qpcr",
        help="design qPCR primers",
        usage="varvamp qpcr -t <threshold> [optional arguments] <alignment> <output dir>"
    )
    parser.add_argument(
        "input",
        nargs=2,
        help="alignment file and dir to write results"
    )
    for par in (SINGLE_parser, TILED_parser, QPCR_parser):
        par.add_argument(
            "-a",
            "--n-ambig",
            metavar="2",
            type=int,
            default=2,
            help="max number of ambiguous characters in a primer"
        )
        par.add_argument(
            "-db",
            "--database",
            help="location of the BLAST db",
            metavar="None",
            type=str,
            default=None
        )
        par.add_argument(
            "-th",
            "--threads",
            help="number of threads",
            metavar="1",
            type=int,
            default=1
        )
        par.add_argument(
            "--name",
            help="name of the scheme",
            metavar="varVAMP",
            type=str,
            default="varVAMP"
        )
        par.add_argument(
            "--compatible-primers",
            metavar="None",
            type=str,
            default=None,
            help="FASTA primer file with which new primers should not form dimers. Sequences >40 nt are ignored. Can significantly increase runtime."
        )

    for par in (SINGLE_parser, TILED_parser):
        par.add_argument(
            "-t",
            "--threshold",
            metavar="",
            type=float,
            default=None,
            help="consensus threshold (0-1) - if not set it will be estimated (higher values result in higher specificity at the expense of found primers)"
        )
        par.add_argument(
            "-ol",
            "--opt-length",
            help="optimal length of the amplicons",
            metavar="1000",
            type=int,
            default=1000
        )
        par.add_argument(
            "-ml",
            "--max-length",
            help="max length of the amplicons",
            metavar="1500",
            type=int,
            default=1500
        )
    TILED_parser.add_argument(
        "-o",
        "--overlap",
        type=int,
        metavar="25",
        default=25,
        help="min overlap of the amplicon inserts"
    )
    SINGLE_parser.add_argument(
        "-n",
        "--report-n",
        type=int,
        metavar="inf",
        default=float("inf"),
        help="report the top n best hits"
    )
    QPCR_parser.add_argument(
        "-pa",
        "--pn-ambig",
        metavar="1",
        type=int,
        default=None,
        help="max number of ambiguous characters in a probe"
    )
    QPCR_parser.add_argument(
        "-t",
        "--threshold",
        required=True,
        type=float,
        help="consensus threshold (0-1) - higher values result in higher specificity at the expense of found primers"
    )
    QPCR_parser.add_argument(
        "-n",
        "--test-n",
        type=int,
        metavar="50",
        default=50,
        help="test the top n qPCR amplicons for secondary structures at the minimal primer temperature"
    )
    QPCR_parser.add_argument(
        "-d",
        "--deltaG",
        type=int,
        metavar="-3",
        default=-3,
        help="minimum free energy (kcal/mol/K) cutoff at the lowest primer melting temperature"
    )
    parser.add_argument(
        "--verbose",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="show varvamp console output"
    )
    parser.add_argument(
        "-v",
        "--version",
        action='version',
        version=f"varvamp {__version__}"
    )
    if len(sysargs) < 1:
        parser.print_help()
        sys.exit(-1)
    else:
        return parser.parse_args(sysargs)


def shared_workflow(args, log_file):
    """
    part of the workflow that is shared by all modes
    """
    # start varvamp
    logging.varvamp_progress(log_file, mode=args.mode)
    # read in alignment and preprocess
    preprocessed_alignment = alignment.preprocess(args.input[0])
    # read in external primer sequences with which new primers should not form dimers
    if args.compatible_primers is not None:
        compatible_primers = primers.parse_primer_fasta(args.compatible_primers)
        if not compatible_primers:
            logging.raise_error(
                "no valid primers found in the provided primer file.\n",
                log_file,
            )
    else:
        compatible_primers = None
    # check alignment length and number of gaps and report if its significantly more/less than expected
    logging.check_alignment_length(preprocessed_alignment, log_file)
    logging.check_gaped_sequences(preprocessed_alignment, log_file)

    # estimate threshold or number of ambiguous bases if args were not supplied
    if args.threshold is None and not args.mode == 'qpcr':
        args.threshold = param_estimation.get_parameters(preprocessed_alignment, args, log_file)
    # set the number of ambiguous chars for qPCR probes to one less than for primers if not given
    if args.mode == "qpcr" and args.pn_ambig is None:
        if args.n_ambig == 0:
            args.pn_ambig = 0
        else:
            args.pn_ambig = args.n_ambig - 1
        with open(log_file, "a") as f:
            print(f"Automatic parameter selection set -pa {args.pn_ambig}.", file=f)

    # check arguments
    logging.raise_arg_errors(args, log_file)

    # config check
    logging.confirm_config(args, log_file)
    logging.varvamp_progress(
        log_file,
        progress=0.1,
        job="Checking config.",
        progress_text="config file passed"
    )

    # clean alignment of gaps
    alignment_cleaned, gaps_to_mask = alignment.process_alignment(
        preprocessed_alignment,
        args.threshold,
    )
    logging.varvamp_progress(
        log_file,
        progress=0.2,
        job="Preprocessing alignment and cleaning gaps.",
        progress_text=f"{len(gaps_to_mask)} gaps with {alignment.calculate_total_masked_gaps(gaps_to_mask)} nucleotides"
    )

    # create consensus sequences
    majority_consensus, ambiguous_consensus = consensus.create_consensus(
        alignment_cleaned,
        args.threshold
    )
    logging.varvamp_progress(
        log_file,
        progress=0.3,
        job="Creating consensus sequences.",
        progress_text=f"length of the consensus is {len(majority_consensus)} nt"
    )

    # generate primer region list
    primer_regions = regions.find_regions(
        ambiguous_consensus,
        args.n_ambig
    )

    potential_primer_regions = regions.mean(primer_regions, majority_consensus)

    if not primer_regions:
        logging.raise_error(
            "no primer regions found. Lower the threshold!",
            log_file,
            exit=True
        )
    logging.varvamp_progress(
        log_file,
        progress=0.4,
        job="Finding primer regions.",
        progress_text=f"{potential_primer_regions} % of the consensus sequence will be evaluated for primers"
    )

    # produce kmers for all primer regions
    kmers = regions.produce_kmers(
        primer_regions,
        majority_consensus
    )
    logging.varvamp_progress(
        log_file,
        progress=0.5,
        job="Digesting into kmers.",
        progress_text=f"{len(kmers)} kmers"
    )

    # find potential primers
    left_primer_candidates, right_primer_candidates = primers.find_primers(
        kmers,
        ambiguous_consensus,
        alignment_cleaned,
        args.threads
    )
    for primer_type, primer_candidates in [("+", left_primer_candidates), ("-", right_primer_candidates)]:
        if not primer_candidates:
            logging.raise_error(
                f"no {primer_type} primers found.\n",
                log_file,
                exit=True
            )
    logging.varvamp_progress(
        log_file,
        progress=0.6,
        job="Filtering for primers.",
        progress_text=f"{len(left_primer_candidates)} fw and {len(right_primer_candidates)} rv potential primers"
    )

    # filter primers against user-provided list of compatible primers, can use multi-processing
    if compatible_primers:
        left_primer_candidates = primers.filter_non_dimer_candidates(
            left_primer_candidates, compatible_primers, args.threads
        )
        right_primer_candidates = primers.filter_non_dimer_candidates(
            right_primer_candidates, compatible_primers, args.threads
        )
        logging.varvamp_progress(
            log_file,
            progress=0.65,
            job="Filtering primers against provided primers.",
            progress_text=f"{len(left_primer_candidates)} fw and {len(right_primer_candidates)} rv primers after filtering"
        )

    return alignment_cleaned, majority_consensus, ambiguous_consensus, primer_regions, left_primer_candidates, right_primer_candidates, potential_primer_regions, compatible_primers


def single_and_tiled_shared_workflow(args, left_primer_candidates, right_primer_candidates, potential_primer_regions, data_dir, log_file):
    """
    part of the workflow shared by the single and tiled mode
    """

    # find best primers and create primer dict
    # depending on the percentage of potential primer regions use high conservation mode
    if potential_primer_regions >= 90:
        all_primers = primers.find_best_primers(left_primer_candidates, right_primer_candidates, high_conservation=True)
        job_text = "Excluding overlapping primers (stringent)."
    else:
        all_primers = primers.find_best_primers(left_primer_candidates, right_primer_candidates, high_conservation=False)
        job_text = "Excluding overlapping primers."
    logging.varvamp_progress(
        log_file,
        progress=0.7,
        job=f"{job_text}",
        progress_text=f"{len(all_primers['+'])} fw and {len(all_primers['-'])} rv primers"
    )

    # find all possible amplicons
    amplicons = scheme.find_amplicons(
        all_primers,
        args.opt_length,
        args.max_length
    )
    if not amplicons:
        logging.raise_error(
            "no amplicons found. Increase the max amplicon length or number of ambiguous bases or lower threshold!\n",
            log_file,
            exit=True
        )
    logging.varvamp_progress(
        log_file,
        progress=0.8,
        job="Finding potential amplicons.",
        progress_text=f"{len(amplicons)} potential amplicons"
    )

    if args.database is not None:
        # create blast query
        query_path = blast.create_BLAST_query(amplicons, data_dir)
        # perform primer blast
        amplicons = blast.primer_blast(
            data_dir,
            args.database,
            query_path,
            amplicons,
            args.max_length,
            args.threads,
            log_file,
            mode="single_tiled"
        )

    return all_primers, amplicons


def single_workflow(args, amplicons, log_file):
    """
    workflow part specific for single mode
    """

    amplicon_scheme = scheme.find_single_amplicons(amplicons, args.report_n)
    logging.varvamp_progress(
        log_file,
        progress=0.9,
        job="Finding amplicons with low penalties.",
        progress_text=f"{len(amplicon_scheme)} amplicons."
    )

    return amplicon_scheme


def tiled_workflow(args, amplicons, left_primer_candidates, right_primer_candidates, all_primers, ambiguous_consensus, log_file):
    """
    part of the workflow specific for the tiled mode
    """

    # create graph
    amplicon_graph = scheme.create_amplicon_graph(amplicons, args.overlap)

    # search for amplicon scheme
    coverage, amplicon_scheme = scheme.find_best_covering_scheme(
        amplicons,
        amplicon_graph
    )

    # evaluate coverage
    # ATTENTION: Genome coverage of the scheme might still change slightly through resolution of primer dimers,
    # but this potential, minor inaccuracy is currently accepted.
    percent_coverage = round(coverage/len(ambiguous_consensus)*100, 2)
    logging.varvamp_progress(
        log_file,
        progress=0.9,
        job="Creating amplicon scheme.",
        progress_text=f"{percent_coverage} % total coverage with {len(amplicon_scheme)} amplicons"
    )
    if percent_coverage < 70:
        logging.raise_error(
            "coverage < 70 %. Possible solutions:\n"
            "\t - lower threshold\n"
            "\t - increase amplicons lengths\n"
            "\t - increase number of ambiguous nucleotides\n"
            "\t - relax primer settings (not recommended)\n",
            log_file
        )

    # check for dimers
    dimers_not_solved, n_initial_dimers = scheme.check_and_solve_heterodimers(
        amplicon_scheme,
        left_primer_candidates,
        right_primer_candidates,
        all_primers,
        args.threads
    )

    # report dimers solve
    if n_initial_dimers > 0 and not dimers_not_solved:
        logging.varvamp_progress(
            log_file,
            progress=0.95,
            job="Trying to solve primer dimers.",
            progress_text=f"all dimers (n={n_initial_dimers}) could be resolved"
        )
    elif dimers_not_solved:
        logging.varvamp_progress(
            log_file,
            progress=0.95,
            job="Trying to solve primer dimers.",
            progress_text=f"{len(dimers_not_solved)}/{n_initial_dimers} dimers could not be resolved"
        )

    return amplicon_scheme, dimers_not_solved


def qpcr_workflow(args, data_dir, alignment_cleaned, ambiguous_consensus, majority_consensus, left_primer_candidates, right_primer_candidates, compatible_primers, log_file):
    """
    part of the workflow specific for the tiled mode
    """
    # find regions for qPCR probes
    probe_regions = regions.find_regions(
        ambiguous_consensus,
        args.pn_ambig
    )
    if not probe_regions:
        logging.raise_error(
            "no regions that fulfill probe criteria! lower threshold or increase number of ambiguous chars in probe\n",
            log_file,
            exit=True
        )

    # digest probe regions
    probe_kmers = regions.produce_kmers(
        probe_regions,
        majority_consensus,
        config.QPROBE_SIZES
    )
    # find potential probes
    qpcr_probes = qpcr.get_qpcr_probes(probe_kmers, ambiguous_consensus, alignment_cleaned, args.threads)
    if not qpcr_probes:
        logging.raise_error(
            "no qpcr probes found\n",
            log_file,
            exit=True
        )
    logging.varvamp_progress(
        log_file,
        progress=0.7,
        job="Finding qPCR probes.",
        progress_text=f"{len(qpcr_probes)} potential qPCR probes"
    )

    # filter primers against non-dimer sequences if provided
    if compatible_primers:
        qpcr_probes = primers.filter_non_dimer_candidates(
            qpcr_probes, compatible_primers, args.threads)
        logging.varvamp_progress(
            log_file,
            progress=0.75,
            job="Filtering probes against provided primers.",
            progress_text=f"{len(qpcr_probes)} potential qPCR probes after filtering"
        )

    # find unique amplicons with a low penalty and an internal probe
    qpcr_scheme_candidates = qpcr.find_qcr_schemes(
        qpcr_probes, left_primer_candidates, right_primer_candidates, majority_consensus, ambiguous_consensus, args.threads
    )
    if not qpcr_scheme_candidates:
        logging.raise_error(
            "no qPCR scheme candidates found. lower threshold or increase number of ambiguous chars in primer and/or probe\n",
            log_file,
            exit=True
        )
    logging.varvamp_progress(
        log_file,
        progress=0.8,
        job="Finding unique amplicons with probe.",
        progress_text=f"{len(qpcr_scheme_candidates)} unique amplicons with internal probe"
    )
    # run blast if db is given
    if args.database is not None:
        # create blast query
        query_path = blast.create_BLAST_query(qpcr_scheme_candidates, data_dir, mode="qpcr")
        # perform primer blast
        qpcr_scheme_candidates = blast.primer_blast(
            data_dir,
            args.database,
            query_path,
            qpcr_scheme_candidates,
            config.QAMPLICON_LENGTH[1],
            args.threads,
            log_file,
            mode="qpcr"
        )
    # test amplicons for deltaG
    final_schemes = qpcr.test_amplicon_deltaG_parallel(qpcr_scheme_candidates, majority_consensus, args.test_n, args.deltaG, args.threads)
    if not final_schemes:
        logging.raise_error(
            "no qPCR amplicon passed the deltaG threshold\n",
            log_file,
            exit=True
        )
    logging.varvamp_progress(
        log_file,
        progress=0.9,
        job="Filtering amplicons for deltaG.",
        progress_text=f"{len(final_schemes)} non-overlapping qPCR schemes that passed deltaG cutoff"
    )
    return probe_regions, final_schemes


def main():
    """
    main varvamp workflow
    """

    # start varVAMP
    args = get_args(sys.argv[1:])
    if not args.verbose:
        sys.stdout = open(os.devnull, 'w')
    start_time = datetime.datetime.now()
    results_dir, data_dir, log_file = logging.create_dir_structure(args.input[1])
    # check if blast is installed
    if args.database is not None:
        blast.check_BLAST_installation(log_file)

    # mode unspecific part of the workflow
    alignment_cleaned, majority_consensus, ambiguous_consensus, primer_regions, left_primer_candidates, right_primer_candidates, potential_primer_regions, compatible_primers = shared_workflow(args, log_file)

    # write files that are shared in all modes
    reporting.write_regions_to_bed(primer_regions, args.name, data_dir)
    reporting.write_alignment(data_dir, alignment_cleaned)
    reporting.write_fasta(data_dir, f"majority_consensus", f"{args.name}_majority_consensus",majority_consensus)
    reporting.write_fasta(results_dir, f"ambiguous_consensus", f"{args.name}_ambiguous_consensus", ambiguous_consensus)

    # Functions called from here on return lists of amplicons that are refined step-wise into final schemes.
    # These lists that are passed between functions and later used for reporting consist of dictionary elemnts,
    # which represent individual amplicons. A minimal amplicon dict could take the form:
    # {
    #     "id": amplicon_name,
    #     "penalty": amplicon_cost,
    #     "length": amplicon_length,
    #     "LEFT": [left primer data],
    #     "RIGHT": [right primer data]
    # }
    # to which different functions may add additional information.

    # SINGLE/TILED mode
    if args.mode == "tiled" or args.mode == "single":
        dimers_not_solved = None
        all_primers, amplicons = single_and_tiled_shared_workflow(
            args,
            left_primer_candidates,
            right_primer_candidates,
            potential_primer_regions,
            data_dir,
            log_file
        )
        if args.mode == "single":
            amplicon_scheme = single_workflow(
                args,
                amplicons,
                log_file
            )
        elif args.mode == "tiled":
            amplicon_scheme, dimers_not_solved = tiled_workflow(
                args,
                amplicons,
                left_primer_candidates,
                right_primer_candidates,
                all_primers,
                ambiguous_consensus,
                log_file,
            )

        # write files
        if args.mode == "tiled":
            # assign amplicon numbers from 5' to 3' along the genome
            amplicon_scheme.sort(key=lambda x: x["LEFT"][1])
        else:
            # make sure amplicons with no off-target products and with low penalties get the lowest numbers
            amplicon_scheme.sort(key=lambda x: (x.get("off_targets", False), x["penalty"]))
        reporting.write_all_primers(data_dir, args.name, all_primers)
        reporting.write_scheme_to_files(
            results_dir,
            amplicon_scheme,
            ambiguous_consensus,
            args.name,
            args.mode,
            log_file,
            dimers_not_solved
        )
        reporting.varvamp_plot(
            results_dir,
            alignment_cleaned,
            primer_regions,
            args.name,
            all_primers=all_primers,
            amplicon_scheme=amplicon_scheme,
        )
        reporting.per_base_mismatch_plot(results_dir, amplicon_scheme, args.threshold, args.name)

    # QPCR mode
    if args.mode == "qpcr":
        probe_regions, final_schemes = qpcr_workflow(
            args,
            data_dir,
            alignment_cleaned,
            ambiguous_consensus,
            majority_consensus,
            left_primer_candidates,
            right_primer_candidates,
            compatible_primers,
            log_file
        )

        # write files
        # make sure amplicons with no off-target products and with low penalties get the lowest numbers
        final_schemes.sort(key=lambda x: (x.get("off_targets", False), x["penalty"]))
        reporting.write_regions_to_bed(probe_regions, args.name, data_dir, "probe")
        reporting.write_qpcr_to_files(results_dir, final_schemes, ambiguous_consensus, args.name, log_file)
        reporting.varvamp_plot(
            results_dir,
            alignment_cleaned,
            primer_regions,
            args.name,
            probe_regions=probe_regions,
            amplicon_scheme=final_schemes
        )
        reporting.per_base_mismatch_plot(results_dir, final_schemes, args.threshold, args.name, mode="QPCR")

    # varVAMP finished
    logging.varvamp_progress(log_file, progress=1, start_time=start_time)
    logging.goodbye_message()



"""
alignment preprocessing
"""

# varVAMP
from varvamp.scripts import config

# LIBS
import numpy as np
from Bio import AlignIO
from Bio.Seq import Seq


def read_alignment(alignment_path):
    """
    read alignment with AlignIO and
    convert to list of lists
    """
    alignment_list = []

    for sequence in AlignIO.read(alignment_path, "fasta"):
        alignment_list.append([sequence.id, str(sequence.seq)])

    return alignment_list


def preprocess(alignment_path):
    """
    force nucleotides to lower and
    back transcribe if its RNA
    """

    preprocessed_alignment = []
    # read alignment
    alignment = read_alignment(alignment_path)

    for sequence in alignment:
        seq = Seq(sequence[1])
        seq = seq.lower()
        if "u" in seq:
            seq = seq.back_transcribe()
        preprocessed_alignment.append([sequence[0], str(seq)])

    return preprocessed_alignment


def clean_gaps(alignment, gaps_to_mask):
    """
    Clean an alignment of large common deletions based on gaps_to_mask.
    gaps_to_mask: list of [start, end] (inclusive), sorted by start.
    """
    cleaned_alignment = []
    gaps_to_mask = sorted(gaps_to_mask, key=lambda x: x[0])

    for seq_id, seq in alignment:
        start = 0
        pieces = []
        # for each seq in the alignment, mask the regions
        for region_start, region_end in gaps_to_mask:
            # mask length for this region
            if (region_end - region_start + 1) >= config.QAMPLICON_DEL_CUTOFF:
                mask = "NN"
            else:
                mask = "N"
            # part before region
            pieces.append(seq[start:region_start])
            # mask for region
            pieces.append(mask)
            # next start is after region
            start = region_end + 1

        # tail after last masked region
        if start < len(seq):
            pieces.append(seq[start:])

        cleaned_alignment.append([seq_id, "".join(pieces)])

    return cleaned_alignment


def process_alignment(preprocessed_alignment, threshold):
    """
    - build an array of shape (n_seq, seq_len)
    - for each column, count how many sequences are '-'
    - mark columns to mask if count > cutoff
    - turn those columns into contiguous regions
    """

    # build char array
    seqs = [seq for seq_id, seq in preprocessed_alignment]
    arr = np.array([list(s) for s in seqs], dtype="U1")
    n_seq, len_seq = arr.shape

    # per-column gap counts
    cols_to_mask = (arr == "-").sum(axis=0) > n_seq * (1 - threshold)

    # convert bool mask into list of (start, end) regions (end inclusive)
    gaps_to_mask = []
    in_gap = False
    start = None
    for i, is_gap in enumerate(cols_to_mask):
        if is_gap and not in_gap:
            in_gap = True
            start = i
        elif not is_gap and in_gap:
            in_gap = False
            gaps_to_mask.append([start, i - 1])
    if in_gap:
        gaps_to_mask.append([start, len_seq - 1])

    if not gaps_to_mask:
        return preprocessed_alignment, []

    alignment_cleaned = clean_gaps(preprocessed_alignment, gaps_to_mask)

    return alignment_cleaned, gaps_to_mask


def calculate_total_masked_gaps(gaps_to_mask):
    """
    calculates the cummulative length of gaps
    that were masked.
    """
    if gaps_to_mask:
        sum_gaps = 0
        for region in gaps_to_mask:
            sum_gaps += region[1] - region[0] + 1
        return sum_gaps
    else:
        return 0



"""
BLAST feature of varVAMP. Checks primers of potential amplicons for hits
against a db and evals if hits are in proximity to potentially
produce off-target amplicons.
"""

# BUILT-INS
import itertools
import multiprocessing
import os
import subprocess
from shutil import which

#LIBS
import pandas as pd

# varVAMP
from varvamp.scripts import config
from varvamp.scripts import logging


def check_BLAST_installation(log_file):
    """
     simple check if BLAST is installed
    """
    if which("blastn") is not None:
        print("\nINFO: BLASTN is installed.")
    else:
        logging.raise_error("BLASTN is not installed", log_file, exit=True)


def create_BLAST_query(amplicons, data_dir, mode="single_tiled"):
    """
    create a query for the BLAST search
    """
    query_path = os.path.join(data_dir, "BLAST_query.fasta")
    if mode == "single_tiled":
        primer_types = ["LEFT", "RIGHT"]
    elif mode == "qpcr":
        primer_types = ["PROBE", "LEFT", "RIGHT"]
    already_written = set()

    with open(query_path, "w") as query:
        for amp in amplicons:
            for primer_type in primer_types:
                name = f"{primer_type}_{amp[primer_type][1]}_{amp[primer_type][2]}"
                if name not in already_written:
                    print(f">{name}\n{amp[primer_type][0]}", file=query)
                    already_written.add(name)
    return query_path


def run_BLAST(query, blast_db, data_dir, n_threads):
    """
    runs a BLAST search on a search query.
    """
    basename = os.path.basename(os.path.normpath(query))
    blast_out = os.path.join(data_dir, f"{basename.strip('.fasta')}_result.tabular")

    blast_command_options = {
        "query": query,
        "db": blast_db,
        "out": blast_out,
        "task": "blastn-short",
        "num_threads": n_threads,
    }
    blast_command_options.update(config.BLAST_SETTINGS)
    args = ['blastn']
    for k, v in blast_command_options.items():
        args.append('-' + k)
        args.append(str(v))
    # run BLAST
    subprocess.run(args, capture_output=True, check=True)

    return blast_out


def parse_and_filter_BLAST_output(blast_out):
    """
    create a BLAST hit database for each primer. filter for mismatches.
    returns a prefiltered pandas df
    """
    columns = ["query",
               "ref",
               "query_len",
               "aln_len",
               "mismatch",
               "gaps",
               "ref_start",
               "ref_end",
               "strand"
               ]

    blast_df = pd.read_table(blast_out, names=columns)
    blast_df = blast_df[
        blast_df["query_len"] - blast_df["aln_len"] + blast_df["mismatch"] + blast_df["gaps"] <=
        blast_df["query_len"]-(round(blast_df["query_len"] * config.BLAST_MAX_DIFF))
        ]
    # removes tabular output
    os.remove(blast_out)

    return blast_df


def check_off_targets(df_amp_primers_sorted, max_length, primers):
    """
    determines off targets in subset blast df
    """
    combinations = list(itertools.combinations(primers, 2))

    for ref in set(df_amp_primers_sorted["ref"]):
        df_ref_subset = df_amp_primers_sorted[df_amp_primers_sorted["ref"] == ref]
        # reindex to remember the correct index later on
        df_ref_subset.reset_index(inplace=True, drop=True)
        # ini the primer search
        indices = []  # remember the index of the subset df
        start = -float("inf")
        for row in df_ref_subset.itertuples():
            # for the current row check if start of the current batch is close enough
            if row[8] - start <= config.BLAST_SIZE_MULTI * max_length:
                indices.append(row[0])
            else:
                # reset start to the next element in the index list if list has more than
                # one element
                if len(indices) > 1:
                    indices.pop(0)
                    start = df_ref_subset.iloc[indices[0]]["ref_start"]
                # else reset to the current row and empty index list
                else:
                    start = row[7]
                    indices = []
            # check if in the df subset is more than one primer
            if len(set(df_ref_subset.iloc[indices]["query"])) <= 1:
                continue
            # subset of df with potential off-targets
            possible_off_targets = df_ref_subset.iloc[indices]
            # check if any two primers in scheme bind different directions
            for combi in combinations:
                direction_fw = set(possible_off_targets[possible_off_targets["query"] == combi[0]]["strand"])
                direction_rv = set(possible_off_targets[possible_off_targets["query"] == combi[1]]["strand"])
                # do we have one primer in subset that binds both directions or are the
                # direction sets different? --> 2 different primers bind on the same chrom,
                # are close enough and are in opposite direction ->[seq]<-
                # >this classifies as an off-target<
                if len(direction_fw) > 1 or len(direction_rv) > 1 or direction_fw != direction_rv:
                    return True
    return False


def predict_non_specific_amplicons_worker(amp, blast_df, max_length, mode):
    """
    Worker function to predict unspecific targets for a single amplicon.
    """
    # get correct primers
    if mode == "single_tiled":
        primer_types = ["LEFT", "RIGHT"]
    elif mode == "qpcr":
        primer_types = ["PROBE", "LEFT", "RIGHT"]
    primers = []
    for primer_type in primer_types:
        primers.append(f"{primer_type}_{amp[primer_type][1]}_{amp[primer_type][2]}")
    # subset df for primers
    df_amp_primers = blast_df[blast_df["query"].isin(primers)]
    # sort by reference and ref start
    df_amp_primers_sorted = df_amp_primers.sort_values(["ref", "ref_start"])
    # check for off-targets for specific primers
    if check_off_targets(df_amp_primers_sorted, max_length, primers):
        amp["off_targets"] = True
    else:
        amp["off_targets"] = False
    return amp


def predict_non_specific_amplicons(amplicons, blast_df, max_length, mode, n_threads):
    """
    Main function to predict unspecific targets within a size range and give
    these primers a high penalty. Uses multiprocessing for parallelization.
    """
    # process amplicons concurrently
    with multiprocessing.Pool(processes=n_threads) as pool:
        annotated_amps = [
            result for result in pool.starmap(
                predict_non_specific_amplicons_worker,
                [(amp, blast_df, max_length, mode) for amp in amplicons]
            ) if result is not None
        ]
    n_off_targets = sum(amp["off_targets"] for amp in annotated_amps)
    return n_off_targets, annotated_amps


def primer_blast(data_dir, db, query_path, amplicons, max_length, n_threads, log_file, mode):
    """
    performs the blast search for the single or tiled workflow
    """
    print("\n#### Starting varVAMP primerBLAST. ####\n")
    print("Running BLASTN...")
    try:
        blast_out = run_BLAST(
            query_path,
            db,
            data_dir,
            n_threads
        )
    except subprocess.CalledProcessError as e:
        logging.raise_error(
            "Failed to run BLAST! Error message: " + str(
                e.stderr
            ) + " Exit code: " + str(
                e.returncode
            ),
            log_file,
            exit=True
        )
    finally:
        # remove query input
        os.remove(query_path)

    blast_df = parse_and_filter_BLAST_output(blast_out)
    print("Predicting non-specific amplicons...")
    n_off_targets, amplicons = predict_non_specific_amplicons(
        amplicons,
        blast_df,
        max_length,
        mode,
        n_threads
    )
    if n_off_targets > 0:
        success_text = f"varVAMP predicted non-specific amplicons:\n\t> {n_off_targets}/{len(amplicons)} amplicons could produce amplicons with the blast db.\n\t> will attempt to avoid them in the final list of amplicons"
    else:
        success_text = f"NO off-target amplicons found with the blast db and a total of {len(amplicons)} amplicons"
    print(success_text)
    with open(log_file, 'a') as f:
        print(
            f"\nBLAST results: {success_text}",
            file=f
        )
    print("\n#### off-target search finished ####\n")

    return amplicons




"""
consensus creation
"""

# BUILT-INS
import collections

# varVAMP
from varvamp.scripts import config


def determine_nucleotide_counts(alignment, idx):
    """
    count the number of each nucleotides at
    an idx of the alignment. return sorted dic.
    handles ambiguous nucleotides in sequences.
    also handles gaps.
    """
    nucleotide_list = []

    # get all nucleotides
    for sequence in alignment:
        nucleotide_list.append(sequence[1][idx])
    # count occurences of nucleotides
    counter = dict(collections.Counter(nucleotide_list))
    # get permutations of an ambiguous nucleotide
    to_delete = []
    temp_dict = {}
    for nucleotide in counter:
        if nucleotide in config.AMBIG_NUCS:
            to_delete.append(nucleotide)
            permutations = config.AMBIG_NUCS[nucleotide]
            adjusted_freq = 1/len(permutations)
            for permutation in permutations:
                if permutation in temp_dict:
                    temp_dict[permutation] += adjusted_freq
                else:
                    temp_dict[permutation] = adjusted_freq
        if nucleotide == "-":
            to_delete.append(nucleotide)

    # drop ambiguous entries and add adjusted freqs to
    if to_delete:
        for i in to_delete:
            counter.pop(i)
        for nucleotide in temp_dict:
            if nucleotide in counter:
                counter[nucleotide] += temp_dict[nucleotide]
            else:
                counter[nucleotide] = temp_dict[nucleotide]

    return dict(sorted(counter.items(), key=lambda x: x[1], reverse=True))


def get_consensus_nucleotides(nucleotide_counts, consensus_cutoff):
    """
    get a list of nucleotides for the consensus seq
    """
    n = 0

    consensus_nucleotides = []
    for nuc in nucleotide_counts:
        n += nucleotide_counts[nuc]
        consensus_nucleotides.append(nuc)
        if n >= consensus_cutoff:
            break

    return consensus_nucleotides


def get_ambiguous_char(nucleotides):
    """
    get ambiguous char from a list of nucleotides
    """
    for ambiguous, permutations in config.AMBIG_NUCS.items():
        if set(permutations) == set(nucleotides):
            return ambiguous


def create_consensus(alignment, threshold):
    """
    build a majority sequence and a sequence that
    has ambiguous chars as determined by the freq
    threshold.
    """

    # ini the consensus seq
    ambiguous_consensus = str()
    majority_consensus = str()

    # define consensus cut-off
    consensus_cutoff = len(alignment)*threshold
    # define length of the consensus from the first seq in alignment
    length_consensus = len(alignment[0][1])

    # built consensus sequences
    for idx in range(length_consensus):
        nucleotide_counts = determine_nucleotide_counts(alignment, idx)
        consensus_nucleotide = get_consensus_nucleotides(
            nucleotide_counts,
            consensus_cutoff
        )
        if len(consensus_nucleotide) > 1:
            amb_consensus_nucleotide = get_ambiguous_char(consensus_nucleotide)
            ambiguous_consensus = ambiguous_consensus + amb_consensus_nucleotide
        else:
            ambiguous_consensus = ambiguous_consensus + consensus_nucleotide[0]

        majority_consensus = majority_consensus + consensus_nucleotide[0]

    return majority_consensus, ambiguous_consensus



"""
This contains all varVAMP parameters.
"""

# List of all known parameters. DO NOT CHANGE!
__all__ = [
    'BLAST_MAX_DIFF', 'BLAST_SETTINGS', 'BLAST_SIZE_MULTI',
    'END_OVERLAP',
    'PCR_DNA_CONC', 'PCR_DNTP_CONC', 'PCR_DV_CONC', 'PCR_MV_CONC',
    'PRIMER_3_PENALTY', 'PRIMER_GC_END', 'PRIMER_GC_PENALTY',
    'PRIMER_GC_RANGE', 'PRIMER_HAIRPIN', 'PRIMER_MAX_BASE_PENALTY',
    'PRIMER_MAX_DIMER_TMP', 'PRIMER_MAX_DIMER_DELTAG', 'PRIMER_MAX_DINUC_REPEATS', 'PRIMER_MAX_POLYX',
    'PRIMER_MIN_3_WITHOUT_AMB', 'PRIMER_PERMUTATION_PENALTY',
    'PRIMER_SIZES', 'PRIMER_SIZE_PENALTY',
    'PRIMER_TMP', 'PRIMER_TM_PENALTY',
    'QAMPLICON_DEL_CUTOFF', 'QAMPLICON_GC', 'QAMPLICON_LENGTH',
    'QPRIMER_DIFF', 'QPROBE_DISTANCE', 'QPROBE_GC_END', 'QPROBE_GC_RANGE',
    'QPROBE_SIZES', 'QPROBE_TEMP_DIFF', 'QPROBE_TMP',
    'NUCS', 'AMBIG_NUCS'
]

# CAN BE CHANGED, DO NOT DELETE

# basic primer parameters
PRIMER_TMP = (56, 63, 60)  # melting temperatur (min, max, opt)
PRIMER_GC_RANGE = (35, 65, 50)  # gc (min, max, opt)
PRIMER_SIZES = (18, 24, 21)  # size (min, max, opt)
PRIMER_MAX_POLYX = 4  # max number of polyX
PRIMER_MAX_DINUC_REPEATS = 4  # max number of polyXY
PRIMER_HAIRPIN = 47  # max melting temp for secondary structure
PRIMER_GC_END = (1, 3)  # min/max GCs in the last 5 bases of the 3' end
PRIMER_MIN_3_WITHOUT_AMB = 3  # min len of 3' without ambiguous charaters
# primer dimer constraints
PRIMER_MAX_DIMER_TMP = 35  # max melting temp for dimers, lower temperature means more stringent filtering
PRIMER_MAX_DIMER_DELTAG = -9000  # max allowed gibbs free energy for dimer formation, higher values mean more stringent filtering
END_OVERLAP = 5  # maximum allowed nt overlap between ends of primers to be considered a dimer

# QPCR parameters
# basic probe parameters
QPROBE_TMP = (64, 70, 67)  # mean 7°C higher than the primer temp
QPROBE_SIZES = (20, 30, 25)
QPROBE_GC_RANGE = (40, 80, 60)
QPROBE_GC_END = (0, 4)
# constraints for amplicon design
QPRIMER_DIFF = 2  # maximal temperature diff of qPCR primers
QPROBE_TEMP_DIFF = (5, 10)  # min/max temp diff between probe and primers
QPROBE_DISTANCE = (4, 15)  # min/max distance to the primer on the same strand
QAMPLICON_LENGTH = (70, 200)  # min/max length of the qPCR amplicon
QAMPLICON_GC = (40, 60)  # GC min/max of the qPCR amplicon
QAMPLICON_DEL_CUTOFF = 4  # consider regions of the alignment for deltaG calculation if they have smaller deletions than cutoff

# PCR parameters
PCR_MV_CONC = 100  # monovalent cations mM
PCR_DV_CONC = 2  # divalent cations mM
PCR_DNTP_CONC = 0.8  # dntp concentration mM
PCR_DNA_CONC = 15  # primer concentration nM

# multipliers for primer and qpcr probe penalties
PRIMER_TM_PENALTY = 2  # temperature penalty
PRIMER_GC_PENALTY = 0.2  # gc penalty
PRIMER_SIZE_PENALTY = 0.5  # size penalty
PRIMER_MAX_BASE_PENALTY = 10  # max base penalty for a primer
PRIMER_3_PENALTY = (32, 16, 8, 4, 2)  # penalties for 3' mismatches
PRIMER_PERMUTATION_PENALTY = 0.1  # penalty for the number of permutations

# BLAST parameters (ref: PrimerBLAST (YE, Jian, et al. Primer-BLAST: a tool to design
# target-specific primers for polymerase chain reaction. BMC bioinformatics, 2012, 13.
# Jg., S. 1-11.)
BLAST_SETTINGS = {  # blast settings for query search
    "outfmt": "6 qseqid sseqid qlen length mismatch gapopen sstart send sstrand",  # do NOT change
    "evalue": 5000,
    "reward": 1,
    "penalty": -1,
    "gapopen": 2,
    "gapextend": 1
}
BLAST_MAX_DIFF = 0.5  # min percent match between primer and BLAST hit (coverage and/or mismatches)
BLAST_SIZE_MULTI = 2  # multiplier for the max_amp size of off targets (in relation to max amp size)

# nucleotide definitions, do NOT change
NUCS = set("atcg")
AMBIG_NUCS = {
    "r": ["a", "g"],
    "y": ["c", "t"],
    "s": ["g", "c"],
    "w": ["a", "t"],
    "k": ["g", "t"],
    "m": ["a", "c"],
    "b": ["c", "g", "t"],
    "d": ["a", "g", "t"],
    "h": ["a", "c", "t"],
    "v": ["a", "c", "g"],
    "n": ["a", "c", "g", "t"]
}



"""
Imports configuration modules (default_config and custom modules).
Represents the imported configurations in a VarVAMPConfig object, does some
basic sanity checks along the way, and merges the default configuration with
custom configuration provided through the $VARVAMP_CONFIG environment variable.

The final VarVAMPConfig instance is what is used as "the" config throughout
varVAMP.
"""

import importlib.machinery
import importlib.util
import os

from varvamp.scripts import default_config


class VarVAMPConfig():
    """
    Parses, validates and stores imported config data.
    """

    __all__ = default_config.__all__

    def __init__(self, config, strict=True):
        defined_attrs = [attr for attr in dir(config) if attr[0] != '_']
        for attr in defined_attrs:
            if attr not in self.__all__:
                # either this is a corrupted default_config,
                # or a custom_config with a wrong param name,
                # or a wrong file was passed as a custom_config by the user.
                raise AttributeError(
                    f'{config.__name__} lists an unknown parameter: "{attr}".'
                    'Please check your VarVAMP config file!'
                )
        for config_option in self.__all__:
            if strict or hasattr(config, config_option):
                # for a custom_config strict will be set to False which will
                setattr(self, config_option, getattr(config, config_option))

    def update(self, other_config):
        for config_option in self.__all__:
            # let any param defined in a custom config overwrite the definition
            # from the default_config
            if hasattr(other_config, config_option):
                setattr(
                    self, config_option, getattr(other_config, config_option)
                )


def load_custom_config(custom_config_path):
    """
    Imports a custom config from a file path and returns its contents as a
    VarVAMPConfig object.
    """

    # based on this stdlib test example:
    # https://github.com/python/cpython/blob/33c1907de2710e7449a38aa27e7b55f274792342/Lib/test/test_asdl_parser.py#L28-L32
    loader = importlib.machinery.SourceFileLoader(
        'custom_config', custom_config_path
    )
    spec = importlib.util.spec_from_loader('custom_config', loader)
    custom_config = importlib.util.module_from_spec(spec)
    loader.exec_module(custom_config)
    return VarVAMPConfig(custom_config, strict=False)


config = VarVAMPConfig(default_config)

custom_config_file = os.getenv('VARVAMP_CONFIG')
if custom_config_file:
    # try to:
    # 1. turn the value of $VARVAMP_CONFIG into an absolute path
    # 2. import from that path
    # 3. update the default configuration with the contents of the custom config
    config.update(
        load_custom_config(
            os.path.normpath(os.path.join(os.getcwd(), custom_config_file))
        )
    )



"""
qPCR probe and amplicon design
"""
import re

# LIBS
import seqfold
import itertools
import multiprocessing
import functools

# varVAMP
from varvamp.scripts import config
from varvamp.scripts import primers


def choose_probe_direction(seq):
    """
    choose the direction of the probe so
    that the probe always has more c than g
    """
    c_count = seq.count("c")
    g_count = seq.count("g")
    direction = "-+"
    if c_count < g_count:
        direction = "-"
    if c_count > g_count:
        direction = "+"

    return direction


def ambiguous_ends(amb_seq):
    """
    determine if kmers have ambiguous ends
    """
    if amb_seq[:1] in config.NUCS and amb_seq[-1] in config.NUCS:
        return False
    else:
        return True


def filter_probe_direction_dependent(seq):
    """
    filter probes for their direction
    """
    return (
            (seq[:1] != "g")  # no 5' g as this leads to quenching
            and (primers.calc_hairpin(seq).tm <= config.PRIMER_HAIRPIN)
            and (config.QPROBE_GC_END[0] <= primers.calc_end_gc(seq) <= config.QPROBE_GC_END[1])
    )


def _process_kmer_batch_probes(ambiguous_consensus, alignment_cleaned, kmers):
    """
    Helper function for multiprocessing: process a batch of kmers for probes.
    Returns probe_candidates dictionary.
    """
    probe_candidates = {}
    probe_idx = 0

    for kmer in kmers:
        if not primers.filter_kmer_direction_independent(kmer[0], config.QPROBE_TMP, config.QPROBE_GC_RANGE, config.QPROBE_SIZES):
            continue
        if ambiguous_ends(ambiguous_consensus[kmer[1]:kmer[2]]):
            continue

        base_penalty = primers.calc_base_penalty(kmer[0], config.QPROBE_TMP, config.QPROBE_GC_RANGE, config.QPROBE_SIZES)
        per_base_mismatches = primers.calc_per_base_mismatches(kmer, alignment_cleaned, ambiguous_consensus)
        permutation_penalty = primers.calc_permutation_penalty(ambiguous_consensus[kmer[1]:kmer[2]])
        direction = choose_probe_direction(kmer[0])

        if "+" in direction:
            if filter_probe_direction_dependent(kmer[0]):
                probe_name = f"PROBE_{probe_idx}_LEFT"
                three_prime_penalty = primers.calc_3_prime_penalty("+", per_base_mismatches)
                probe_candidates[probe_name] = [kmer[0], kmer[1], kmer[2],
                                                base_penalty + permutation_penalty + three_prime_penalty,
                                                per_base_mismatches, direction]
                probe_idx += 1
        if "-" in direction:
            if filter_probe_direction_dependent(primers.rev_complement(kmer[0])):
                probe_name = f"PROBE_{probe_idx}_RIGHT"
                three_prime_penalty = primers.calc_3_prime_penalty("-", per_base_mismatches)
                probe_candidates[probe_name] = [primers.rev_complement(kmer[0]), kmer[1], kmer[2],
                                                base_penalty + permutation_penalty + three_prime_penalty,
                                                per_base_mismatches, direction]
                probe_idx += 1

    return probe_candidates


def get_qpcr_probes(kmers, ambiguous_consensus, alignment_cleaned, num_processes):
    """
    Find potential qPCR probes using multiprocessing.
    """

    # Convert kmers set to list for batching
    kmers = list(kmers)

    # Split kmers into batches
    batch_size = max(1, int(len(kmers) / num_processes))
    batches = [kmers[i:i + batch_size] for i in range(0, len(kmers), batch_size)]

    # Prepare arguments for each dimer
    callable_f = functools.partial(
        _process_kmer_batch_probes,
        ambiguous_consensus, alignment_cleaned
    )
    with multiprocessing.Pool(processes=num_processes) as pool:
        results = pool.map(callable_f, batches)

    # Aggregate results and re-index probe names
    probe_candidates = {}
    probe_idx = 0
    for batch_probes in results:
        if batch_probes is None:
            continue
        for probe_name, probe_data in batch_probes.items():
            # Extract direction from original probe name
            direction = "LEFT" if "LEFT" in probe_name else "RIGHT"
            new_probe_name = f"PROBE_{probe_idx}_{direction}"
            probe_candidates[new_probe_name] = probe_data
            probe_idx += 1

    # Sort by penalty
    probe_candidates = dict(sorted(probe_candidates.items(), key=lambda x: x[1][3]))

    return probe_candidates


def flanking_primer_subset(primer_list, primer_direction, probe):
    """
    subset for primers flanking the probe and sort by penalty
    """
    subset = []

    if primer_direction == "+":
        window_start = probe[1] - config.QAMPLICON_LENGTH[1] + len(probe[0])
        window_stop = probe[1]
    else:
        window_start = probe[2]
        window_stop = probe[2] + config.QAMPLICON_LENGTH[1] - len(probe[0])
    for primer in primer_list:
        if window_start < primer[1] and primer[2] < window_stop:
            subset.append(primer)

    # sort by penalty and start if same penalty
    subset.sort(key=lambda x: (x[3], x[1]))

    return subset


def hardfilter_amplicon(majority_consensus, left_primer, right_primer):
    """
    hardfilter possible amplicon for length, gc and if there are Ns
    (indicating deletions in amp) or n (no folding analysis possible)
    """
    amplicon_length = right_primer[2] - left_primer[1]
    amplicon_seq = majority_consensus[left_primer[1]:right_primer[2]]
    # check for the right amplicon length
    return (
            (config.QAMPLICON_LENGTH[0] <= amplicon_length <= config.QAMPLICON_LENGTH[1])
            and (config.QAMPLICON_GC[0] <= primers.calc_gc(amplicon_seq) <= config.QAMPLICON_GC[1])
            and "NN" not in amplicon_seq
    )


def dimer_in_combinations(right_primer, left_primer, probe, ambiguous_consensus):
    """
    checks if primers cause dimers and if combinations of primers/probe including all permutations form dimers
    """

    forms_structure = False

    # first check if there are dimers between the two flanking primers
    if primers.is_dimer(left_primer[0], right_primer[0]):
        return True
    # for the probe check all permutations and possible overhangs to ensure
    # that none of the primers could cause unspecific probe binding.
    # first get all permutations
    probe_per = primers.get_permutations(ambiguous_consensus[probe[1]:probe[2]])
    left_per = primers.get_permutations(ambiguous_consensus[left_primer[1]:left_primer[2]])
    right_per = primers.get_permutations(ambiguous_consensus[right_primer[1]:right_primer[2]])
    # then check all permutations
    for combination in [(probe_per, left_per), (probe_per, right_per)]:
        for oligo1, oligo2 in itertools.product(*combination):
            if primers.is_dimer(oligo1, oligo2):
                forms_structure = True
                break
        # break also outer loop because we found an unwanted structure in one of the permutations
        # (either dimer formation or a too long overlap at the ends of the primer)
        if forms_structure:
            break

    return forms_structure


def assess_amplicons(left_subset, right_subset, qpcr_probes, probe, majority_consensus, ambiguous_consensus):
    """
    assess if a potential amplicon is a qPCR scheme for a specific probe and return the one with the lowest penalty
    """

    primer_combinations = ()
    amplicon_found = False

    # consider a combination of flanking primers if ...
    for left_primer in left_subset:
        for right_primer in right_subset:
            # ... the amplicon is large enough and is in gc range, ...
            if not hardfilter_amplicon(majority_consensus, left_primer, right_primer):
                continue
            # ... the probe is close enough to the primer on the same strand
            if "LEFT" in probe:
                if not qpcr_probes[probe][1] in range(
                        left_primer[2] + config.QPROBE_DISTANCE[0],
                        left_primer[2] + config.QPROBE_DISTANCE[1]
                ):
                    continue
            elif "RIGHT" in probe:
                if not right_primer[1] in range(
                        qpcr_probes[probe][2] + config.QPROBE_DISTANCE[0],
                        qpcr_probes[probe][2] + config.QPROBE_DISTANCE[1]

                ):
                    continue
            # ... the primer temps do not differ too much, ...
            primer_temps = (primers.calc_temp(right_primer[0]), primers.calc_temp(left_primer[0]))
            if abs(primer_temps[0] - primer_temps[1]) > config.QPRIMER_DIFF:
                continue
            # ... the probe has a higher temp than the primers and ...
            probe_temp = primers.calc_temp(qpcr_probes[probe][0])
            if not all(
                    [config.QPROBE_TEMP_DIFF[0] <= probe_temp - x <= config.QPROBE_TEMP_DIFF[1] for x in primer_temps]):
                continue
            # .... all combination of oligos do not form dimers or overhangs.
            if dimer_in_combinations(right_primer, left_primer, qpcr_probes[probe], ambiguous_consensus):
                continue
            # append to list and break as this is the primer combi
            # with the lowest penalty (primers are sorted by penalty)
            amplicon_found = True
            break
        # break also the outer loop
        if amplicon_found:
            primer_combinations = (left_primer, right_primer)
            break

    return primer_combinations


def find_single_qpcr_scheme(left_primer_candidates, right_primer_candidates, qpcr_probes,
                            majority_consensus, ambiguous_consensus, probe):
    """
    Find a qPCR scheme for a single probe.
    """

    probe_name, probe_data = probe

    # Generate flanking subsets within the worker process
    left_subset = flanking_primer_subset(left_primer_candidates, "+", probe_data)
    right_subset = flanking_primer_subset(right_primer_candidates, "-", probe_data)

    if not left_subset or not right_subset:
        return probe_name, None

    primer_combination = assess_amplicons(
        left_subset, right_subset, qpcr_probes, probe_name,
        majority_consensus, ambiguous_consensus
    )

    return probe_name, primer_combination


def find_qcr_schemes(qpcr_probes, left_primer_candidates, right_primer_candidates,
                     majority_consensus, ambiguous_consensus, num_processes):
    """
    Find final qPCR schemes using multiprocessing to evaluate probes in parallel.
    Probes are sorted by penalty, ensuring optimal probe selection.
    """
    qpcr_scheme_candidates = []
    found_amplicons = []
    amplicon_nr = -1

    # Prepare arguments for parallel processing - pass full primer lists
    batch_size = max(1, int(len(qpcr_probes) / num_processes))
    callable_f = functools.partial(
        find_single_qpcr_scheme,
        left_primer_candidates, right_primer_candidates, qpcr_probes, majority_consensus, ambiguous_consensus
    )

    # Process probes in parallel
    with multiprocessing.Pool(processes=num_processes) as pool:
        results = pool.map(callable_f, qpcr_probes.items(), chunksize=batch_size)

    # Aggregate results in original probe order (sorted by penalty)
    for probe_name, primer_combination in results:
        if not primer_combination:
            continue
        if primer_combination in found_amplicons:
            continue

        amplicon_nr += 1
        found_amplicons.append(primer_combination)
        qpcr_scheme_candidates.append({
            "id": f"AMPLICON_{amplicon_nr}",
            "penalty": qpcr_probes[probe_name][3] + primer_combination[0][3] + primer_combination[1][3],
            "PROBE": qpcr_probes[probe_name],
            "LEFT": primer_combination[0],
            "RIGHT": primer_combination[1]
        })

    # Sort by total penalty
    qpcr_scheme_candidates.sort(key=lambda x: x["penalty"])

    return qpcr_scheme_candidates


def process_single_amplicon_deltaG(majority_consensus, amplicon):
    """
    Process a single amplicon to test its deltaG and apply filtering.
    This function will be called concurrently by multiple threads.
    """
    seq = majority_consensus[amplicon["LEFT"][1]:amplicon["RIGHT"][2]]
    seq = seq.replace("N", "")
    seq = seq.replace("n", "")
    # check if the amplicon overlaps with an amplicon that was previously
    # found and had a high enough deltaG
    min_temp = min((primers.calc_temp(amplicon["LEFT"][0]),
                    primers.calc_temp(amplicon["RIGHT"][0])))
    # calculate deltaG at the minimal primer temp
    amplicon["deltaG"] = seqfold.dg(seq, min_temp)

    return amplicon


def test_amplicon_deltaG_parallel(qpcr_schemes_candidates, majority_consensus, n_to_test, deltaG_cutoff, n_processes):
    """
    Test all amplicon deltaGs for the top n hits at the lowest primer temperature
    and filters if they fall below the cutoff. Multiple processes are used
    for processing amplicons in parallel.
    """
    final_amplicons = []

    # Create a list of the first n amplicon tuples for processing
    # The list is sorted first on whether offset targets were predicted for the amplicon,
    # then by penalty. This ensures that amplicons with offset targets are always considered last
    amplicons = list(sorted(qpcr_schemes_candidates, key=lambda x: (x.get("offset_targets", False), x["penalty"])))[:n_to_test]
    # process amplicons concurrently
    batch_size = max(1, int(n_to_test / n_processes))
    callable_f = functools.partial(
        process_single_amplicon_deltaG,
        majority_consensus
    )
    with multiprocessing.Pool(processes=n_processes) as pool:
        results = pool.map(callable_f, amplicons, chunksize=batch_size)
    # Process the results
    retained_ranges = []
    for amp in results:
        # check if the amplicon overlaps with an amplicon that was previously
        # found and had a high enough deltaG
        if amp["deltaG"] <= deltaG_cutoff:
            continue
        amp_range = range(amp["LEFT"][1], amp["RIGHT"][2])
        overlaps_retained = False
        for r in retained_ranges:
            if amp_range.start < r.stop and r.start < amp_range.stop:
                overlaps_retained = True
                break
        if not overlaps_retained:
            final_amplicons.append(amp)
            retained_ranges.append(amp_range)

    return final_amplicons



"""
estimate varVAMP threshold if not given
"""

# libs
import numpy as np

# varVAMP
from varvamp.scripts import config


def calculate_frequencies(preprocessed_alignment):
    """
    calculate individual frequencies for a, t, c, g, and gaps at each position
    returns a 2D numpy array where each column is a position and rows are [a, c, t, g, -]
    """
    # convert alignment to numpy array
    alignment_array = np.array([list(seq.lower()) for _, seq in preprocessed_alignment])
    num_sequences, sequence_length = alignment_array.shape

    # calculate frequencies
    frequencies = np.zeros((5, sequence_length))
    nucleotides = ['a', 'c', 't', 'g', '-']

    for i in range(sequence_length):
        column = alignment_array[:, i]
        for nuc_idx, nuc in enumerate(nucleotides):
            frequencies[nuc_idx, i] = np.count_nonzero(column == nuc) / num_sequences

    return np.max(frequencies[0:4, :], axis=0), frequencies[4, :]


def calculate_distances(highest_freq, deletion_freq, threshold):
    """
    calc the distance for each nuc freq to the prior
    nuc freq that fell below the cutoff
    """

    current_dis = 0
    previous_dis = 0
    all_dis = []

    for idx, (freq, del_freq) in enumerate(zip(highest_freq, deletion_freq)):
        if freq < threshold:
            current_dis = 0
        # ignore gaps -> they will be excluded later and might skew distance calc
        if del_freq > 1-threshold:
            continue
        current_dis += 1
        if current_dis <= previous_dis or idx == len(highest_freq)-1:
            all_dis.append(previous_dis)

        previous_dis = current_dis

    return all_dis


def get_parameters(preprocessed_alignment, args, log_file):
    """
    give an estimate for number of ambiguous chars and/or threshold
    writes to log file

    How it works:

    - calculate a np array with frequencies per position for ATCG and gaps (-)
    - increment for different threshold starting at 0.1
    - for each increment calculate:
        - the lengths of conseq. nucleotides that are >= threshold
        - skip all positions with gaps >= 1 - threshold
        - this results in an array of stretch lengths
        - Then check for each stretch if it could be considered as a primer region:
            e.g. 4 ambiguous bases:
                stretches: [1,1,1,1,6,10,5,3,1,1]
                - first check for the current stretch if the sum of the next n stretches (with n being
                the n ambiguous bases + 1) is larger than the minimal length of a primer
                - this would result that the with stars marked positions would pass:
                [1,1,1,1*,6*,10*,5,3,1,1]
        - now calculate the distance between the current start of a stretch and the previous stop
        - if this is larger than the minimal amplicon length - exit the optimization
        - reset the threshold to second prior iteration to make it more robust (allow more primer regions) as
        sometimes the optimization would fail with just one iteration

    Manual optimization can be beneficial in some cases.
    """
    # set coverage to max
    args.threshold = 0.1
    # calc freqs
    highest_frequencies, deletion_frequency = calculate_frequencies(preprocessed_alignment)
    # write to log
    with open(log_file, 'a') as f:
        print(f"AUTOMATIC THRESHOLD SELECTION\n", file=f)
        print(f"-t\tmaximum non-covered region", file=f)
        # calculate distance between passing potential primer regions
        previous_stop = 0
        while args.threshold < 1:
            distances = calculate_distances(highest_frequencies, deletion_frequency, args.threshold)
            max_distance_between_passing, previous_passing = 0, 0
            # check if the distance between potential primer regions is not larger than:
            # args.opt_length - 2 * args.overlap
            for idx, dis in enumerate(distances):
                if sum(distances[idx:idx + 1 + args.n_ambig]) >= config.PRIMER_SIZES[1]:
                    # the stretch start in the gap-excluded alignment is the sum of all prior distances including the current
                    # minus the distance of the current stretch
                    stretch_start = sum(distances[:idx])
                    # then the distance between the prior stop and current start is calculated
                    current_dis = stretch_start - previous_stop
                    # and the max is updated if necessary
                    if max_distance_between_passing < current_dis:
                        max_distance_between_passing = current_dis
                    # update previous stop position
                    previous_stop = stretch_start + distances[idx]
            # write each iteration to log
            print(round(args.threshold, 2), max_distance_between_passing, sep="\t", file=f)
            # check if the distance is acceptable
            distance_threshold = args.opt_length - 2 * args.overlap if args.mode == 'tiled' else args.opt_length
            if max_distance_between_passing < distance_threshold:
                # never exceed 0.99
                if args.threshold < 0.99:
                    args.threshold += 0.01
                else:
                    break
            # or reset to the param of the two previous iterations
            else:
                args.threshold -= 0.02
                break
        print(f"Automatic parameter selection set -t {round(args.threshold, 2)} at -a {args.n_ambig}.", file=f)

    return round(args.threshold, 2)



"""
varVAMP logging and raising errors
"""

# BUILT-INS
import sys
import os
import shutil
import datetime
import random
import statistics
import re

# varVAMP
from varvamp.scripts import config
from varvamp import __version__


def create_dir_structure(dir_path):
    """
    create output folders and log file
    """
    cwd = os.getcwd()
    results_dir = os.path.join(cwd, dir_path)
    data_dir = os.path.join(results_dir, "data/")
    # create folders
    if not os.path.exists(results_dir):
        os.makedirs(results_dir)
    else:
        shutil.rmtree(results_dir)
        os.makedirs(results_dir)
    os.makedirs(data_dir)

    log_file = os.path.join(results_dir, "varvamp_log.txt")

    return results_dir, data_dir, log_file


def varvamp_progress(log_file, mode=None, start_time=None, progress=0, job="", progress_text=""):
    """
    progress bar, main progress logging and folder creation
    """
    barLength = 40
    block = int(round(barLength * progress))

    if progress == 0:
        print(
            "\nStarting \033[31m\033[1mvarVAMP ◥(ºwº)◤\033[0m primer design\n",
            flush=True
        )
        with open(log_file, 'w') as f:
            print(f"VARVAMP log \n\nMODE = {mode}\n", file=f)
    else:
        if progress == 1:
            stop_time = datetime.datetime.now() - start_time
            progress_text = f"all done \n\n\rvarVAMP {__version__} finished in {stop_time.total_seconds()} sec!\n{datetime.datetime.now()}"
            job = "Finalizing output."
        print(
            "\rJob:\t\t " + job + "\nProgress: \t [{0}] {1}%".format("█" * block + "-" * (barLength - block),
                                                                     progress * 100) + "\t" + progress_text,
            flush=True
        )
        with open(log_file, 'a') as f:
            print(
                f"\rJob:\t {job} \nResult:\t {progress_text}",
                file=f
            )


def raise_error(message, log_file, exit=False):
    """
    raises warnings or errors, writes to log
    """
    # print to log
    with open(log_file, 'a') as f:
        if exit:
            print(f"ERROR: {message}", file=f)
        else:
            print(f"WARNING: {message}", file=f)
    # print to console
    if exit:
        sys.exit(f"\n\033[31m\033[1mERROR:\033[0m {message}")
    else:
        print(f"\033[31m\033[1mWARNING:\033[0m {message}")


def raise_arg_errors(args, log_file):
    """
    checks arguments for non-valid input and raises warnings
    """
    # threshold error
    if args.threshold > 1 or args.threshold < 0:
        raise_error(
            "threshold can only be between 0-1",
            log_file,
            exit=True
        )
    if args.n_ambig < 0:
        raise_error(
            "number of ambiguous chars can not be negative",
            log_file,
            exit=True
        )
    if args.n_ambig > 4:
        raise_error(
            "high number of ambiguous nucleotides in primer leads to a high "
            "degeneracy. Consider reducing.",
            log_file
        )
    if args.threads < 1:
        raise_error(
            "number of threads cannot be smaller than 1.",
            log_file,
            exit=True
        )
    if args.mode in ("tiled", "single"):
        if args.opt_length > args.max_length:
            raise_error(
                "optimal length can not be higher than the maximum amplicon length.",
                log_file,
                exit=True
            )
        if args.opt_length < 0 or args.max_length < 0:
            raise_error(
                "amplicon lengths can not be negative.",
                log_file,
                exit=True
            )
    # SINGLE specific warnings
    if args.mode == "single":
        if args.report_n < 1:
            raise_error(
                "number of reported amplicons cannot be below 1.",
                log_file,
                exit=True
            )
    # TILED specific warnings
    if args.mode == "tiled":
        if args.opt_length < 200 or args.max_length < 200:
            raise_error(
                "your amplicon lengths might be to small. Consider increasing",
                log_file
            )
        if args.overlap < 0:
            raise_error(
                "overlap size can not be negative.",
                log_file,
                exit=True
            )
        if args.overlap < 10:
            raise_error(
                "small overlaps might hinder downstream analyses. Consider increasing.",
                log_file
            )
        if args.overlap > args.max_length / 2 - config.PRIMER_SIZES[0] * 2:
            raise_error(
                "min overlap must be lower than your maximum length / 2 - 2 * min primer length.",
                log_file,
                exit=True
            )
        if args.overlap > args.opt_length:
            raise_error(
                "overlap can not be higher than your optimal length.",
                log_file,
                exit=True
            )
        if args.overlap > args.opt_length / 2 - config.PRIMER_SIZES[0] * 2:
            raise_error(
                "your min overlap is lower than your optimal length / 2 - 2 * min primer length. This reduces how well varvamps will find overlapping amplicons. Consider decreasing.",
                log_file
            )
    # QPCR specific warnings
    if args.mode == "qpcr":
        if args.pn_ambig < 0:
            raise_error(
                "number of ambiguous characters in the qPCR probe cannot be negative.",
                log_file,
                exit=True
            )
        if args.pn_ambig > args.n_ambig:
            raise_error(
                "the degeneracy of qPCR probes should not be higher than that of qPCR primers to retain specificity",
                log_file,
                exit=True
            )
        if args.test_n <= 0:
            raise_error(
                "if the number of tested qPCR amplicons is 0 or lower, no amplicons will be reported.",
                log_file,
                exit=True
            )

        if args.test_n > 200:
            raise_error(
                "checking the deltaG of amplicons is computationally intensive. setting this higher than 200 will likely take a bit to compute.",
                log_file
            )
        if args.pn_ambig > 2:
            raise_error(
                "setting the degeneracy of probes higher than 2 can result in variable probe design. Consider reducing!",
                log_file
            )
        if args.deltaG > 0:
            raise_error(
                "setting the deltaG cutoff larger than 0 is not recommended!",
                log_file
            )


def check_alignment_length(preprocessed_alignment, log_file):
    """
    checks the sequence lengths of the alignment and reports a warning
    if the length is larger or smaller than the mean +- 3std of all seqs
    and at least 1 % diff to mean
    """

    all_seq_length, all_names = [], []
    # count the length
    for seq in preprocessed_alignment:
        all_names.append(seq[0])
        all_seq_length.append(len(seq[1].strip("-")))
    # clac mean and std
    mean_len, mean_std = statistics.mean(all_seq_length), statistics.stdev(all_seq_length)
    # check for each seq if it is larger or smaller than the mean +-3std
    # and is at least 1 % smaller or larger than the alignment mean
    # (otherwise varvamp will report very small differences).
    smaller_warning, larger_warning = [], []
    for name, length in zip(all_names, all_seq_length):
        # consider variation of sequence lengths and check if it is at least 2 % deviation
        if length <= mean_len - 3 * mean_std and length <= mean_len - mean_len * 0.02:
            smaller_warning.append(f"{name} ({length} nt)\n")
        elif length >= mean_len + 3 * mean_std and length >= mean_len + mean_len * 0.02:
            larger_warning.append(f"{name} ({length} nt)\n")
    # raise warning for non-empty lists
    for warning, length_type in zip([larger_warning, smaller_warning], ["larger", "smaller"]):
        if not warning:
            continue
        warning = "".join(warning)
        raise_error(
            f"The following sequences are considerably {length_type} than the alignment mean ({round(mean_len)} nt) and might cause alignment trimming:\n{warning}",
            log_file,
            exit=False
        )


def check_gaped_sequences(preprocessed_alignment, log_file):
    """
    checks the number of gaps in each sequence of the alignment and reports a warning
    if the number of gaps is larger than the mean + 3std
    """
    number_of_gaps = {}

    for seq in preprocessed_alignment:
        # find all gaps for all sequences with regular expression -{min}
        number_of_gaps[seq[0]] = len(list(re.finditer(r"-+", seq[1])))

    # clac mean and std
    mean_gaps, mean_gaps_std = statistics.mean(number_of_gaps.values()), statistics.stdev(number_of_gaps.values())

    warning = []

    for name, n_gaps in number_of_gaps.items():
        if n_gaps < mean_gaps - 3 * mean_gaps_std:
            warning.append(f"{name} ({n_gaps} gaps)\n")

    if warning:
        raise_error(
            f"The following sequences contain less gaps than the alignment mean ({round(mean_gaps)} gaps) and might overproportionally gap the alignment:\n{"".join(warning)}",
            log_file,
            exit=False
        )


def confirm_config(args, log_file):
    """
    checks the config. raises error and warnings
    if nececarry. writes settings to log
    """
    error = False

    # check if all variables exists
    all_vars = [
        # arg independent all modes
        (
            "PRIMER_TMP",
            "PRIMER_GC_RANGE",
            "PRIMER_SIZES",
            "PRIMER_HAIRPIN",
            "PRIMER_MAX_POLYX",
            "PRIMER_MAX_DINUC_REPEATS",
            "PRIMER_GC_END",
            "PRIMER_MAX_DIMER_TMP",
            "PRIMER_MAX_DIMER_DELTAG",
            "PRIMER_MIN_3_WITHOUT_AMB",
            "PCR_MV_CONC",
            "PCR_DV_CONC",
            "PCR_DNTP_CONC",
            "PCR_DNA_CONC",
            "PRIMER_TM_PENALTY",
            "PRIMER_GC_PENALTY",
            "PRIMER_SIZE_PENALTY",
            "PRIMER_MAX_BASE_PENALTY",
            "PRIMER_3_PENALTY",
            "PRIMER_PERMUTATION_PENALTY",
            "END_OVERLAP"
        ),
        # arg independent QPCR mode
        (
            "QPROBE_TMP",
            "QPROBE_SIZES",
            "QPROBE_GC_RANGE",
            "QPROBE_GC_END",
            "QPRIMER_DIFF",
            "QPROBE_TEMP_DIFF",
            "QPROBE_DISTANCE",
            "QAMPLICON_LENGTH",
            "QAMPLICON_GC",
            "QAMPLICON_DEL_CUTOFF"
        ),
        (
            "BLAST_MAX_DIFF",
            "BLAST_SIZE_MULTI",
        )
    ]

    for tup in all_vars:
        for var in tup:
            if var not in vars(config):
                raise_error(
                    f"{var} does not exist in config!",
                    log_file
                )
                error = True
    # exit if variables are not defined
    if error:
        raise_error(
            "config is missing parameters. Look at the above warnings!",
            log_file,
            exit=True
        )
    # confirm tuples with 3 elements
    for tup_type, tup in [("primer temp", config.PRIMER_TMP), ("primer gc", config.PRIMER_GC_RANGE),
                          ("primer size", config.PRIMER_SIZES), ("probe temp", config.QPROBE_TMP),
                          ("probe size", config.QPROBE_SIZES), ("probe gc", config.QPROBE_GC_RANGE)]:
        if len(tup) != 3:
            raise_error(
                f"{tup_type} tuple has to have the form (min, max, opt)!",
                log_file
            )
            error = True
        if tup[0] > tup[1]:
            raise_error(
                f"min {tup_type} should not exceed max {tup_type}!",
                log_file
            )
            error = True
        if tup[0] > tup[2]:
            raise_error(
                f"min {tup_type} should not exceed opt {tup_type}!",
                log_file
            )
            error = True
        if tup[2] > tup[1]:
            raise_error(
                f"opt {tup_type} should not exeed max {tup_type}!",
                log_file
            )
            error = True
        if any(map(lambda var: var < 0, tup)):
            raise_error(
                f"{tup_type} can not contain negative values!",
                log_file
            )
            error = True
    # confirm tuples with two elements
    for tup_type, tup in [("primer gc at the 3'", config.PRIMER_GC_END),
                          ("probe and primer temp diff", config.QPROBE_TEMP_DIFF),
                          ("distance probe to primer", config.QPROBE_DISTANCE),
                          ("qpcr amplicon length", config.QAMPLICON_LENGTH), ("qpcr amplicon gc", config.QAMPLICON_GC),
                          ("probe gc at the 3'", config.PRIMER_GC_END)]:
        if len(tup) != 2:
            raise_error(
                f"{tup_type} tuple has to have the form (min, max)!",
                log_file
            )
            error = True
        if tup[0] > tup[1]:
            raise_error(
                f"min {tup_type} should not exceed max {tup_type}!",
                log_file
            )
            error = True
        if any(map(lambda var: var < 0, tup)):
            raise_error(
                f"{tup_type} can not contain negative values!",
                log_file
            )
            error = True
    # check single values that cannot be negative
    non_negative_var = [
        ("min number of 3 prime nucleotides without ambiguous nucleotides", config.PRIMER_MIN_3_WITHOUT_AMB),
        ("monovalent cation concentration", config.PCR_MV_CONC),
        ("divalent cation concentration", config.PCR_DV_CONC),
        ("dNTP concentration", config.PCR_DNTP_CONC),
        ("primer temperature penalty", config.PRIMER_TM_PENALTY),
        ("primer gc penalty", config.PRIMER_GC_PENALTY),
        ("primer size penalty", config.PRIMER_SIZE_PENALTY),
        ("max base penalty", config.PRIMER_MAX_BASE_PENALTY),
        ("primer permutation penalty", config.PRIMER_PERMUTATION_PENALTY),
        ("qpcr flanking primer difference", config.QPRIMER_DIFF),
        ("end overlap", config.END_OVERLAP),
        ("qpcr deletion size still considered for deltaG calculation", config.QAMPLICON_DEL_CUTOFF),
        ("maximum difference between primer and blast db", config.BLAST_MAX_DIFF),
        ("multiplier of the maximum length for non-specific amplicons", config.BLAST_SIZE_MULTI),
    ]
    for var_type, var in non_negative_var:
        if var < 0:
            raise_error(
                f"{var_type} can not be negative!",
                log_file
            )
            error = True
    if any(map(lambda var: var < 0, config.PRIMER_3_PENALTY)):
        raise_error(
            "3' penalties can not be zero!",
            log_file
        )
        error = True
    # exit if variables are not properly defined
    if error:
        raise_error(
            "config has flaws. Look at the above warnings!",
            log_file,
            exit=True
        )
    # specific warnings
    if config.PRIMER_MAX_POLYX > 5:
        raise_error(
            "max polyx > 5 is not recommended.",
            log_file,
        )
    if config.PRIMER_MAX_DINUC_REPEATS > 5:
        raise_error(
            "max di-nucleotide repeats > 5 is not recommended.",
            log_file,
        )
    if config.PRIMER_HAIRPIN < 0:
        raise_error(
            "decreasing hairpin melting temp to negative values will influence successful primer search!",
            log_file
        )
    if config.PRIMER_MAX_DIMER_TMP < 21:
        raise_error(
            "there is no need to set max dimer melting temp below room temperature.",
            log_file
        )
    if config.PRIMER_MAX_DIMER_DELTAG > -6000:
        raise_error(
            "primer interactions with deltaG values higher than -6000 are generally considered unproblematic.",
            log_file
        )

    if config.PRIMER_MAX_BASE_PENALTY < 8:
        raise_error(
            "decreasing the base penalty will filter out more primers.",
            log_file
        )
    if config.PRIMER_GC_END[0] > 3 or config.QPROBE_GC_END[0] > 3:
        raise_error(
            "large GC clamps will results in too high 3'end stability",
            log_file
        )
    if config.PRIMER_GC_END[1] > 5 or config.QPROBE_GC_END[1] > 5:
        raise_error(
            "only the last 5 nucleotides of the 3' end are considered for GC 3'end calculation.",
            log_file
        )
    # specific errors
    if config.PRIMER_MAX_POLYX < 1:
        raise_error(
            "max polyx cannot be lower than 1.",
            log_file,
            exit=True
        )
    if config.PRIMER_MAX_DINUC_REPEATS < 1:
        raise_error(
            "max di-nucleotide repeats cannot be lower than 1.",
            log_file,
            exit=True
        )
    if config.BLAST_MAX_DIFF > 1:
        raise_error(
            "max difference to the blast db should be between 0-1.",
            log_file,
            exit=True
        )
    if config.BLAST_SIZE_MULTI < 1:
        raise_error(
            "off-targets should be considered at least in the range of the maximal amplicon length.",
            log_file,
            exit=True
        )
    # confirm proper BLAST settings in dictionary
    if not isinstance(config.BLAST_SETTINGS, dict):
        raise_error(
            "BLAST settings have to be in a dictionary format!",
            log_file,
            exit=True
        )
    if config.BLAST_SETTINGS["outfmt"] != "6 qseqid sseqid qlen length mismatch gapopen sstart send sstrand":
        raise_error(
            "output format (outfmt) of BLAST settings is not correct",
            log_file,
            exit=True
        )

    # check all dict values
    dict_values = ["evalue", "reward", "penalty", "gapopen", "gapextend"]

    for blast_setting in dict_values:
        if blast_setting not in config.BLAST_SETTINGS:
            raise_error(
                f"blast setting ({blast_setting}) is missing in BLAST dict",
                log_file,
                exit=True
            )
        if not isinstance(config.BLAST_SETTINGS[blast_setting], int):
            raise_error(
                f"blast setting ({blast_setting}) has to be an integer",
                log_file,
                exit=True
            )

    # write all settings to file
    var_dic = vars(config)
    with open(log_file, 'a') as f:
        print(
            "\nARG SETTINGS\n",
            f"THRESHOLD = {args.threshold}",
            f"PRIMER_ALLOWED_N_AMB = {args.n_ambig}",
            sep="\n",
            file=f
        )
        if args.mode in ("tiled", "single"):
            print(
                f"AMPLICON_OPT_LENGTH = {args.opt_length}",
                f"AMPLICON_MAX_LENGTH = {args.max_length}",
                sep="\n",
                file=f
            )
            if args.database is not None:
                print(
                    f"BLAST_DATABASE = {args.database}",
                    f"N_THREADS = {args.threads}",
                    sep="\n",
                    file=f
                )
        if args.mode == "tiled":
            print(
                f"MIN_OVERLAP = {args.overlap}",
                sep="\n",
                file=f
            )
        if args.mode == "single":
            print(
                f"REPORT_N_AMPLICONS = {args.report_n}",
                sep="\n",
                file=f
            )
        if args.mode == "qpcr":
            print(
                f"PROBE_ALLOWED_N_AMB = {args.pn_ambig}",
                f"TEST_DELTAG_N_AMPLICONS = {args.test_n}",
                f"DELTAG_CUTOFF = {args.deltaG}",
                sep="\n",
                file=f
            )
        print(
            "\nCONFIG SETTINGS\n",
            sep="\n",
            file=f
        )
        for var in all_vars[0]:
            print(f"{var} = {var_dic[var]}", file=f)
        if args.mode in ("tiled", "single"):
            if args.database is not None:
                for var in all_vars[2]:
                    print(f"{var} = {var_dic[var]}", file=f)
        if args.mode == "qpcr":
            for var in all_vars[1]:
                print(f"{var} = {var_dic[var]}", file=f)
        print("\nPROGRESS", file=f)


def goodbye_message():
    """
    This is what happens when I am bored...
    """

    messages = [
        "Thank you. Come again.",
        ">Placeholder for your advertisement<",
        "Ciao cacao!",
        "And now lets pray to the PCR gods.",
        "**bibobibobop** task finished",
        "Thank you for traveling with varVAMP.",
        "This message will self-destruct in 0 seconds.... PUFF!",
        "What? Already finished? Errr I mean, thanks for using varVAMP!",
        "One primer design to rule them all.",
        "42",
        "Hope to see you again!",
        "Wohooooo everything worked somehow!",
        "Reminder: Get yourself a cup of coffee!",
        "Barba non facit philosophum.",
        "Task failed successfully.",
        "Never gonna give you up, never gonna let you down.",
        "Have you tried turning it off and on again?",
        "Just try it. PCR is magic!",
        "One goat was sacrificed for this primer design to work.",
        "You seem trustworthy. Here's a cookie!",
        "Owwww yeah, primers done!",
        "These primers were designed without AI assistance.",
        "Research is fun (if you ignore the pipetting).",
        "Balance your primers, balance your life.",
        "Keep calm and PCR on.",
        "In silico we trust.",
        "May the primers be with you.",
        "Designing primers like a boss!",
        "Primer design completed. Time for a break!",
        "Eureka! Your primers are ready.",
        "Look, I am your primer scheme.",
        "Quod erat demonstrandum.",
        "Miau?",
        "This is an automated message informing you that you are awsome.",
        "Why was the negative-sense virus angry at the positive-sense virus?\nBecause he was left stranded!",
        "If you see this message twice, you are an experienced user.",
        "No one expects the spanish inquisition!",
        "Primer design you must."
    ]
    print(f"\n{random.choice(messages)}")



"""
amplicon search
"""

# BUILT-INS
import heapq
import math
import multiprocessing
import functools

# varVAMP
from varvamp.scripts import config, primers


def construct_graph(nodes, init_graph):
    """
    This method makes sure that the graph is symmetrical, but sets the costs
    for nodes in the reverse direction to infinity to make sure dijkstra
    never goes to an amplicon that is in the wrong direction.
    """
    graph = {}
    for node in nodes:
        graph[node] = {}

    graph.update(init_graph)
    for node, neighbors in graph.items():
        for neighbor in neighbors.keys():
            if graph[neighbor].get(node, False) is False:
                graph[neighbor][node] = (float("infinity"), 0)
    return graph


class Graph(object):
    """
    a graph class
    """

    def __init__(self, nodes, init_graph):
        self.nodes = nodes
        self.graph = construct_graph(nodes, init_graph)

    def get_nodes(self):
        """
        Returns the nodes of the graph.
        """
        return self.nodes

    def get_neighbors(self, node):
        """
        Returns the neighbors of a node.
        """
        neighbors = []
        for out_node in self.nodes:
            if self.graph[node].get(out_node, False) is not False:
                neighbors.append(out_node)
        return neighbors

    def value(self, node1, node2):
        """
        Returns the value of an edge between two nodes.
        """
        return self.graph[node1][node2]

def find_amplicons(all_primers, opt_len, max_len):
    """
    finds all possible amplicons, creates a dictionary
    """
    amplicon_number = 0
    amplicons = []

    for left_name in all_primers["+"]:
        left_primer = all_primers["+"][left_name]
        for right_name in all_primers["-"]:
            right_primer = all_primers["-"][right_name]
            amplicon_length = right_primer[2] - left_primer[1]
            if not opt_len <= amplicon_length <= max_len:
                continue
            if primers.is_dimer(left_primer[0], right_primer[0]):
                continue
            # calculate length dependend amplicon costs as the cumulative primer
            # penalty multiplied by the e^(fold length of the optimal length).
            amplicon_costs = (right_primer[3] + left_primer[3])*math.exp(amplicon_length/opt_len)
            amplicon_name = "AMPLICON_"+str(amplicon_number)
            amplicons.append(
                {
                    "id": amplicon_name,
                    "penalty": amplicon_costs,
                    "length": amplicon_length,
                    "LEFT": left_primer + [left_name],
                    "RIGHT": right_primer + [right_name],
                }
            )
            amplicon_number += 1
    return amplicons


def has_qualifying_overlap(current_amplicon, next_amplicon, min_overlap):
    """
    check if two amplicons overlap sufficiently to connect them in the graph
    """
    # connect amplicons if they sufficiently overlap because:
    # ... the start of next amplicon lies in the second half of the prior amplicon
    if next_amplicon["LEFT"][1] < current_amplicon["LEFT"][1] + current_amplicon["length"] / 2:
        return False
    # ... the stop of the left primer of the next amplicon does not lie in the minimum amplicon insert
    if next_amplicon["LEFT"][2] > current_amplicon["RIGHT"][1] - min_overlap:
        return False
    # ... half of the next amplicon does not overlap with the previous amplicon --> enough space for a
    # further amplicon that lies in the second half next amplicon and cannot overlap with a primer of the
    # current amplicon
    if next_amplicon["RIGHT"][2] <= current_amplicon["RIGHT"][2] + next_amplicon["length"] / 2:
        return False

    return True


def create_amplicon_graph(amplicons, min_overlap):
    """
    creates the amplicon graph.
    """
    # ini graph and vertices
    amplicon_graph = {}
    nodes = []

    for current_amplicon in amplicons:
        # remember all vertices
        amplicon_id = current_amplicon["id"]
        nodes.append(amplicon_id)
        for next_amplicon in amplicons:
            if not has_qualifying_overlap(current_amplicon, next_amplicon, min_overlap):
                continue
            # --> write to graph
            if amplicon_id not in amplicon_graph:
                amplicon_graph[amplicon_id] = {
                    next_amplicon["id"]: (
                        next_amplicon.get("off_targets", False),
                        next_amplicon["penalty"]
                    )
                }
            else:
                amplicon_graph[amplicon_id][next_amplicon["id"]] = (
                    next_amplicon.get("off_targets", False),
                    next_amplicon["penalty"]
                )

    # return a graph object
    return Graph(nodes, amplicon_graph)


def dijkstra_algorithm(graph, start_node):
    """
    implementation of the dijkstra algorithm
    """

    previous_nodes = {}
    shortest_path = {node: (float('infinity'), 0) for node in graph.get_nodes()}
    shortest_path[start_node] = (0, 0)

    nodes_to_test = [((0, 0), start_node)]

    while nodes_to_test:
        current_distance, current_node = heapq.heappop(nodes_to_test)
        if current_distance > shortest_path[current_node]:
            continue
        for neighbor in graph.get_neighbors(current_node):
            off_targets, base_penalty = graph.value(current_node, neighbor)
            distance = (
                current_distance[0] + off_targets,
                current_distance[1] + base_penalty
            )
            # Only consider this new path if it's a better path
            if not distance < shortest_path[neighbor]:
                continue
            shortest_path[neighbor] = distance
            previous_nodes[neighbor] = current_node
            heapq.heappush(nodes_to_test, (distance, neighbor))

    return previous_nodes, shortest_path


def get_end_node(previous_nodes, shortest_path, amplicons):
    """
    get the target node with the lowest penalty costs out of all
    nodes that have the same maximum end position
    """
    stop_nucleotide, possible_end_nodes = 0, {}

    for node in previous_nodes.keys():
        # check if a node has a larger stop -> empty dict and set new
        # best stop nucleotide
        amplicon_stop = amplicons[node]["RIGHT"][2]
        if amplicon_stop > stop_nucleotide:
            possible_end_nodes = {node: shortest_path[node]}
            stop_nucleotide = amplicon_stop
        # if nodes have the same stop nucleotide, add to dictionary
        elif amplicon_stop == stop_nucleotide:
            possible_end_nodes[node] = shortest_path[node]

    # return the end node with the lowest penalty costs
    return min(possible_end_nodes.items(), key=lambda x: x[1])


def get_min_path(previous_nodes, start_node, target_node):
    """
    get the min path from the start to stop node from the
    previosuly calculated shortest path
    """
    path = []
    node = target_node

    while node != start_node:
        path.append(node)
        node = previous_nodes[node]
    # Add the start node manually
    path.append(start_node)

    # return the inverse list
    return path[::-1]


def create_scheme(amplicon_path, amplicons_by_id):
    """
    creates the final tiled-amplicon scheme
    """
    amplicon_scheme = []

    for pool in (0, 1):
        for amp_id in amplicon_path[pool::2]:
            amplicons_by_id[amp_id]["pool"] = pool
            amplicon_scheme.append(amplicons_by_id[amp_id])

    return amplicon_scheme


def find_best_covering_scheme(amplicons, amplicon_graph):
    """
    this brute forces the amplicon scheme search until the largest
    coverage with the minimal costs is achieved.
    """
    # ini
    best_coverage = 0
    max_stop = max(amplicons, key=lambda x: x["RIGHT"][2])["RIGHT"][2]
    lowest_costs = (float('infinity'),)
    # a dict for fast access to amplicons by their ID
    amps_by_id = {amp["id"]: amp for amp in amplicons}
    for amplicon in amplicons:
        # if the currently best coverage + start nucleotide of the currently tested amplicon
        # is smaller than the maximal stop nucleotide there might be a better amplicon
        # scheme that covers more of the genome
        if amplicon["LEFT"][1] + best_coverage <= max_stop:
            previous_nodes, shortest_path = dijkstra_algorithm(amplicon_graph, amplicon["id"])
            # only continue if there are previous_nodes
            if previous_nodes:
                target_node, costs = get_end_node(previous_nodes, shortest_path, amps_by_id)
                coverage = amps_by_id[target_node]["RIGHT"][2] - amplicon["LEFT"][1]
                # if the new coverage is larger, go for the larger coverage
                if coverage > best_coverage:
                    best_start_node = amplicon["id"]
                    best_target_node = target_node
                    best_previous_nodes = previous_nodes
                    lowest_costs = costs
                    best_coverage = coverage
                # if the coverages are identical, go for the lowest costs
                elif coverage == best_coverage:
                    if costs < lowest_costs:
                        best_start_node = amplicon["id"]
                        best_target_node = target_node
                        best_previous_nodes = previous_nodes
                        lowest_costs = costs
                        best_coverage = coverage
            else:
                # check if the single amplicon has the largest coverage so far
                coverage = amplicon["length"]
                if coverage > best_coverage:
                    best_start_node = amplicon["id"]
                    best_previous_nodes = previous_nodes
                    lowest_costs = (
                        amplicon.get("off_targets", False), amplicon["penalty"])
                    best_coverage = coverage
        # no need to check more, the best covering amplicon scheme was found and
        # has the minimal costs compared to the schemes with the same coverage
        else:
            break

    if best_previous_nodes:
        amplicon_path = get_min_path(best_previous_nodes, best_start_node, best_target_node)
    else:
        # if no previous nodes are found but the single amplicon results in the largest
        # coverage - return as the best scheme
        amplicon_path = [best_start_node]

    return best_coverage, create_scheme(amplicon_path, amps_by_id)


def test_scheme_for_dimers(amplicon_scheme):
    """
    test the lowest-cost scheme for primer dimers
    """

    primer_dimers = []
    non_dimers = {amp["pool"]:set() for amp in amplicon_scheme}
    # write all primer sequences in the respective pools -->
    # these primers should not be violated by primer switching
    # and primers are only switched later if no primer dimers
    # with the existing 'good' scheme are created
    for amp in amplicon_scheme:
        non_dimers[amp["pool"]].add(amp["LEFT"][0])
        non_dimers[amp["pool"]].add(amp["RIGHT"][0])
    for pool in non_dimers:
        # test the primer dimers only within the respective pools
        tested_primers = []
        for amp_index, amp in enumerate(amplicon_scheme):
            if amp["pool"] != pool:
                continue
            for primer in ["LEFT", "RIGHT"]:
                # remember where the currrent primer was in the scheme
                # store the amplicon's index in the scheme, the current primer's original name and its details
                current_primer = (amp_index, amp[primer][-1], amp[primer][:-1])
                current_seq = current_primer[2][0]
                for tested in tested_primers:
                    tested_seq = tested[2][0]
                    if not primers.is_dimer(current_seq, tested_seq):
                        continue
                    primer_dimers.append((current_primer, tested))
                    non_dimers[pool].discard(current_seq)
                    non_dimers[pool].discard(tested_seq)
                # and remember all tested primers
                tested_primers.append(current_primer)

    # report both dimers and non-dimers
    return primer_dimers, non_dimers


def get_overlapping_primers(dimer, left_primer_candidates, right_primer_candidates):
    """
    get overlapping primers of a primer dimer pair that have been previously
    excluded. returns list of list with possible primers for each primer
    in primer dimer.
    """

    overlapping_primers = []
    # test each primer in dimer
    for amp_index, primer_name, primer in dimer:
        overlapping_primers_temp = []
        # as switching could violate overlap criteria,
        # only consider primers that overlap in the left half (LEFT primers)
        # or right half (RIGHT primers) respectively, however this can result in slightly
        # longer amplicons than allowed.
        half_length = int((primer[2] - primer[1]) / 2)
        if "RIGHT" in primer_name:
            overlap_set = set(range(primer[1] + half_length, primer[2]))
            primers_to_test = right_primer_candidates
        else:
            overlap_set = set(range(primer[1], primer[1] + half_length))
            primers_to_test = left_primer_candidates
        # and check this list for all primers that overlap
        for potential_new in primers_to_test:
            primer_positions = list(range(potential_new[1], potential_new[2]))
            if not any(x in primer_positions for x in overlap_set):
                continue
            overlapping_primers_temp.append((amp_index, primer_name, potential_new))

        overlapping_primers.append(overlapping_primers_temp)

    return overlapping_primers


def test_overlaps_for_dimers(overlapping_primers, non_dimers):
    """
    test all possible overlapping primers against each other for dimers
    and return the first pair that doesn't form a dimer with each other
    and with all non-dimer forming primers in the pool.
    """
    for first_overlap in overlapping_primers[0]:
        if any(primers.is_dimer(seq, first_overlap[2][0]) for seq in non_dimers):
            continue
        for second_overlap in overlapping_primers[1]:
            if any(primers.is_dimer(seq, second_overlap[2][0]) for seq in non_dimers):
                continue
            if not primers.is_dimer(first_overlap[2][0], second_overlap[2][0]):
                return [first_overlap, second_overlap]


def _solve_single_dimer(amplicon_scheme, left_primer_candidates, right_primer_candidates, non_dimers_all_pools, dimer):
    """
    Helper function for multiprocessing: solve a single dimer independently.
    Returns (amp_index, primer_name, new_primer) tuples or empty list if no solution.
    """
    pool = amplicon_scheme[dimer[0][0]]["pool"]
    non_dimers = non_dimers_all_pools[pool]

    overlapping_primers = get_overlapping_primers(dimer, left_primer_candidates, right_primer_candidates)
    new_primers = test_overlaps_for_dimers(overlapping_primers, non_dimers)

    return new_primers if new_primers else []


def check_and_solve_heterodimers(amplicon_scheme, left_primer_candidates, right_primer_candidates, all_primers, num_processes):
    """
    check scheme for heterodimers, try to find
    new primers that overlap and replace the existing ones.
    Uses multiprocessing to solve dimers in parallel.
    """
    primer_dimers, non_dimers_all_pools = test_scheme_for_dimers(amplicon_scheme)
    n_initial_dimers = len(primer_dimers)

    if not primer_dimers:
        return [], 0

    # Prepare arguments for each dimer
    callable_f = functools.partial(
        _solve_single_dimer,
    amplicon_scheme, left_primer_candidates, right_primer_candidates, non_dimers_all_pools
    )

    # Solve dimers in parallel
    with multiprocessing.Pool(processes=num_processes) as pool:
        results = pool.map(callable_f, primer_dimers)

    # Apply all solutions to the scheme
    for new_primers in results:
        if new_primers:
            for amp_index, primer_name, primer in new_primers:
                # overwrite in final scheme
                # ATTENTION: doesn't update the amplicon penalty currently
                # This is ok only because that value isn't used after and doesn't get reported anywhere.
                if "LEFT" in primer_name:
                    strand = "+"
                    amplicon_scheme[amp_index]["LEFT"] = primer + [primer_name]
                else:
                    strand = "-"
                    amplicon_scheme[amp_index]["RIGHT"] = primer + [primer_name]
                amplicon_scheme[amp_index]["length"] = amplicon_scheme[amp_index]["RIGHT"][2] - amplicon_scheme[amp_index]["LEFT"][1]
                # and in all primers
                all_primers[strand][primer_name] = primer
    # get remaining dimers in the revised scheme and add pool identifier for reporting
    remaining_primer_dimers, _ = test_scheme_for_dimers(amplicon_scheme)
    primer_dimers = [
        (amplicon_scheme[primer1[0]]["pool"], primer1, primer2)
        for primer1, primer2 in remaining_primer_dimers
    ]

    return primer_dimers, n_initial_dimers


def find_single_amplicons(amplicons, n):
    """
    find non-overlapping amplicons with low penalties
    from all found amplicons. only for the SINGLE mode.
    """
    # sort amplicons
    sorted_amplicons = sorted(amplicons, key=lambda x: (x.get("off_targets", False), x["penalty"]))
    to_retain = []
    retained_ranges = []
    # find lowest non-overlapping
    for amp in sorted_amplicons:
        overlaps_retained = False
        amp_range = range(amp["LEFT"][1], amp["RIGHT"][2])
        for r in retained_ranges:
            if amp_range.start < r.stop and r.start < amp_range.stop:
                overlaps_retained = True
                break
        if not overlaps_retained:
            retained_ranges.append(amp_range)
            to_retain.append(amp)
            if len(to_retain) == n:
                break

    return to_retain
